#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for nxos_prefix_lists
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: nxos_prefix_lists
short_description: Prefix-Lists resource module.
description:
- This module manages prefix-lists configuration on devices running Cisco NX-OS.
version_added: 2.4.0
notes:
- Tested against NX-OS 9.3.6.
- Unsupported for Cisco MDS
- This module works with connection C(network_cli) and C(httpapi).
author: Nilashish Chakraborty (@NilashishC)
options:
  running_config:
    description:
    - This option is used only with state I(parsed).
    - The value of this option should be the output received from the NX-OS device
      by executing the command B(show running-config | section '^ip(.*) prefix-list').
    - The state I(parsed) reads the configuration from C(running_config) option and
      transforms it into Ansible structured data as per the resource module's argspec
      and the value is then returned in the I(parsed) key within the result.
    type: str
  config:
    description: A list of prefix-list configuration.
    type: list
    elements: dict
    suboptions:
      afi:
        description:
        - The Address Family Identifier (AFI) for the prefix-lists.
        type: str
        choices: ["ipv4", "ipv6"]
      prefix_lists:
        description: List of prefix-list configurations.
        type: list
        elements: dict
        suboptions:
          name:
            description: Name of the prefix-list.
            type: str
          description:
            description: Description of the prefix list
            type: str
          entries:
            description: List of configurations for the specified prefix-list
            type: list
            elements: dict
            suboptions:
              sequence:
                description: Sequence Number.
                type: int
              action:
                description: Prefix-List permit or deny.
                type: str
                choices: ["permit", "deny"]
              prefix:
                description: IP or IPv6 prefix in A.B.C.D/LEN or A:B::C:D/LEN format.
                type: str
              eq:
                description: Exact prefix length to be matched.
                type: int
              ge:
                description: Minimum prefix length to be matched.
                type: int
              le:
                description: Maximum prefix length to be matched.
                type: int
              mask:
                description: Explicit match mask.
                type: str
  state:
    description:
    - The state the configuration should be left in.
    - Refer to examples for more details.
    - With state I(replaced), for the listed prefix-lists,
      sequences that are in running-config but not in the task are negated.
    - With state I(overridden), all prefix-lists that are in running-config but
      not in the task are negated.
    - Please refer to examples for more details.
    type: str
    choices:
    - merged
    - replaced
    - overridden
    - deleted
    - parsed
    - gathered
    - rendered
    default: merged
"""
EXAMPLES = """
# Using merged

# Before state:
# -------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# nxos-9k-rdo#

- name: Merge the provided configuration with the existing running configuration
  cisco.nxos.nxos_prefix_lists:
    config:
      - afi: ipv4
        prefix_lists:
          - name: AllowPrefix
            description: allows engineering IPv4 networks
            entries:
              - sequence: 10
                action: permit
                prefix: 192.0.2.0/23
                eq: 24
              - sequence: 20
                action: permit
                prefix: 198.51.100.128/26
          - name: DenyPrefix
            description: denies lab IPv4 networks
            entries:
              - sequence: 20
                action: deny
                prefix: 203.0.113.0/24
                le: 25

      - afi: ipv6
        prefix_lists:
          - name: AllowIPv6Prefix
            description: allows engineering IPv6 networks
            entries:
              - sequence: 8
                action: permit
                prefix: "2001:db8:400::/38"
              - sequence: 20
                action: permit
                prefix: "2001:db8:8000::/35"
                le: 37

# Task output
# -------------
# before: []
#
# commands:
#   - "ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks"
#   - "ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38"
#   - "ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37"
#   - "ip prefix-list AllowPrefix description allows engineering IPv4 networks"
#   - "ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24"
#   - "ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26"
#   - "ip prefix-list DenyPrefix description denies lab IPv4 networks"
#   - "ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25"
#
# after:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix

# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

# Using replaced

# Before state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Replace prefix-lists configurations of listed prefix-lists with provided configurations
  cisco.nxos.nxos_prefix_lists:
    config:
      - afi: ipv4
        prefix_lists:
          - name: AllowPrefix
            description: allows engineering IPv4 networks
            entries:
              - sequence: 10
                action: permit
                prefix: 203.0.113.64/27

              - sequence: 30
                action: permit
                prefix: 203.0.113.96/27
          - name: AllowPrefix2Stub
            description: allow other engineering IPv4 network
    state: replaced

# Task output
# -------------
# before:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# commands:
#   - "no ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24"
#   - "ip prefix-list AllowPrefix seq 10 permit 203.0.113.64/27"
#   - "ip prefix-list AllowPrefix seq 30 permit 203.0.113.96/27"
#   - "no ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26"
#   - "ip prefix-list AllowPrefix2Stub description allow other engineering IPv4 network"
#
# after:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 203.0.113.64/27
#           - sequence: 30
#             action: permit
#             prefix: 203.0.113.96/27
#          name: AllowPrefix
#       - description: allow other engineering IPv4 network
#         name: AllowPrefix2Stub
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 203.0.113.64/27
# ip prefix-list AllowPrefix seq 30 permit 203.0.113.96/27
# ip prefix-list AllowPrefix2Stub description allow other engineering IPv4 network
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

# Using overridden

# Before state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Override all prefix-lists configuration with provided configuration
  cisco.nxos.nxos_prefix_lists: &id003
    config:
      - afi: ipv4
        prefix_lists:
          - name: AllowPrefix
            description: allows engineering IPv4 networks
            entries:
              - sequence: 10
                action: permit
                prefix: 203.0.113.64/27

              - sequence: 30
                action: permit
                prefix: 203.0.113.96/27
          - name: AllowPrefix2Stub
            description: allow other engineering IPv4 network
    state: overridden

# Task output
# -------------
# before:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# commands:
#   - "no ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24"
#   - "ip prefix-list AllowPrefix seq 10 permit 203.0.113.64/27"
#   - "ip prefix-list AllowPrefix seq 30 permit 203.0.113.96/27"
#   - "no ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26"
#   - "ip prefix-list AllowPrefix2Stub description allow other engineering IPv4 network"
#   - "no ip prefix-list DenyPrefix"
#   - "no ipv6 prefix-list AllowIPv6Prefix"
#
# after:
#   - afi: ipv4
#     prefix_lists:
#       - name: AllowPrefix
#         description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 203.0.113.64/27
#
#           - sequence: 30
#             action: permit
#             prefix: 203.0.113.96/27
#       - name: AllowPrefix2Stub
#         description: allow other engineering IPv4 network
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 203.0.113.64/27
# ip prefix-list AllowPrefix seq 30 permit 203.0.113.96/27
# ip prefix-list AllowPrefix2Stub description allow other engineering IPv4 network

# Using deleted to delete a all prefix lists for an AFI

# Before state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Delete all prefix-lists for an AFI
  cisco.nxos.nxos_prefix_lists:
    config:
      - afi: ipv4
    state: deleted
  register: result

# Task output
# -------------
# before:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# commands:
#   - "no ip prefix-list AllowPrefix"
#   - "no ip prefix-list DenyPrefix"
#
# after:
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

# Using deleted to delete a single prefix-list

# Before state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Delete a single prefix-list
  cisco.nxos.nxos_prefix_lists:
    config:
      - afi: ipv4
        prefix_lists:
          - name: AllowPrefix
    state: deleted

# Task output
# -------------
# before:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# commands:
#   - "no ip prefix-list AllowPrefix"
#
# after:
#   - afi: ipv4
#     prefix_lists:
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

# Using deleted to delete all prefix-lists from the device

# Before state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Delete all prefix-lists
  cisco.nxos.nxos_prefix_lists:
    state: deleted

# Task output
# -------------
# before:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
#
# commands:
#   - "no ip prefix-list AllowPrefix"
#   - "no ip prefix-list DenyPrefix"
#   - "no ipv6 prefix-list AllowIPv6Prefix"
#
# after: []
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | section 'ip(.*) prefix-list'
# nxos-9k-rdo#

# Using rendered

- name: Render platform specific configuration lines with state rendered (without connecting to the device)
  cisco.nxos.nxos_prefix_lists: &id001
    config:
      - afi: ipv4
        prefix_lists:
          - name: AllowPrefix
            description: allows engineering IPv4 networks
            entries:
              - sequence: 10
                action: permit
                prefix: 192.0.2.0/23
                eq: 24
              - sequence: 20
                action: permit
                prefix: 198.51.100.128/26
          - name: DenyPrefix
            description: denies lab IPv4 networks
            entries:
              - sequence: 20
                action: deny
                prefix: 203.0.113.0/24
                le: 25

      - afi: ipv6
        prefix_lists:
          - name: AllowIPv6Prefix
            description: allows engineering IPv6 networks
            entries:
              - sequence: 8
                action: permit
                prefix: "2001:db8:400::/38"
              - sequence: 20
                action: permit
                prefix: "2001:db8:8000::/35"
                le: 37
    state: rendered

# Task Output (redacted)
# -----------------------
# rendered:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix

# Using parsed

# parsed.cfg
# ------------
# ip prefix-list AllowPrefix description allows engineering IPv4 networks
# ip prefix-list AllowPrefix seq 10 permit 192.0.2.0/23 eq 24
# ip prefix-list AllowPrefix seq 20 permit 198.51.100.128/26
# ip prefix-list DenyPrefix description denies lab IPv4 networks
# ip prefix-list DenyPrefix seq 20 deny 203.0.113.0/24 le 25
# ipv6 prefix-list AllowIPv6Prefix description allows engineering IPv6 networks
# ipv6 prefix-list AllowIPv6Prefix seq 8 permit 2001:db8:400::/38
# ipv6 prefix-list AllowIPv6Prefix seq 20 permit 2001:db8:8000::/35 le 37

- name: Parse externally provided prefix-lists configuration
  register: result
  cisco.nxos.nxos_prefix_lists:
    running_config: "{{ lookup('file', './parsed.cfg') }}"
    state: parsed

# Task output (redacted)
# -----------------------
# parsed:
#   - afi: ipv4
#     prefix_lists:
#       - description: allows engineering IPv4 networks
#         entries:
#           - sequence: 10
#             action: permit
#             prefix: 192.0.2.0/23
#             eq: 24
#           - sequence: 20
#             action: permit
#             prefix: 198.51.100.128/26
#         name: AllowPrefix
#       - description: denies lab IPv4 networks
#         entries:
#           - sequence: 20
#             action: deny
#             prefix: 203.0.113.0/24
#             le: 25
#         name: DenyPrefix
#
#   - afi: ipv6
#     prefix_lists:
#       - description: allows engineering IPv6 networks
#         entries:
#           - sequence: 8
#             action: permit
#             prefix: "2001:db8:400::/38"
#           - sequence: 20
#             action: permit
#             prefix: "2001:db8:8000::/35"
#             le: 37
#         name: AllowIPv6Prefix
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.prefix_lists.prefix_lists import (
    Prefix_listsArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.prefix_lists.prefix_lists import (
    Prefix_lists,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Prefix_listsArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Prefix_lists(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



# Copyright (c) 2014, Chris Church <chris@ninemoreminutes.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    name: csh
    short_description: C shell (/bin/csh)
    description:
      - When you have no other option than to use csh
    extends_documentation_fragment:
      - shell_common
'''

from shlex import quote as shlex_quote
from ansible.plugins.shell import ShellBase


class ShellModule(ShellBase):

    # Common shell filenames that this plugin handles
    COMPATIBLE_SHELLS = frozenset(('csh', 'tcsh'))
    # Family of shells this has.  Must match the filename without extension
    SHELL_FAMILY = 'csh'

    # commonly used
    ECHO = 'echo'
    COMMAND_SEP = ';'

    # How to end lines in a python script one-liner
    _SHELL_EMBEDDED_PY_EOL = '\\\n'
    _SHELL_REDIRECT_ALLNULL = '>& /dev/null'
    _SHELL_AND = '&&'
    _SHELL_OR = '||'
    _SHELL_SUB_LEFT = '"`'
    _SHELL_SUB_RIGHT = '`"'
    _SHELL_GROUP_LEFT = '('
    _SHELL_GROUP_RIGHT = ')'

    def env_prefix(self, **kwargs):
        ret = []
        # All the -u options must be first, so we process them first
        ret += ['-u %s' % k for k, v in kwargs.items() if v is None]
        ret += ['%s=%s' % (k, shlex_quote(str(v))) for k, v in kwargs.items() if v is not None]
        return 'env %s' % ' '.join(ret)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Phil Schwartz <schwartzmx@gmail.com>
# Copyright: (c) 2015, Trond Hindenes
# Copyright: (c) 2015, Hans-Joachim Kliemeck <git@kliemeck.de>
# Copyright: (c) 2020, Håkon Heggernes Lerring <hakon@lerring.no>
# Copyright: (c) 2023, Jordan Pitlor <jordan@pitlor.dev>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_acl
short_description: Set file/directory/registry/certificate permissions for a system user or group
description:
- Add or remove rights/permissions for a given user or group for the specified
  file, folder, registry key or AppPool identifies.
options:
  path:
    description:
    - The path to the file or directory.
    type: str
    required: yes
  user:
    description:
    - User or Group to add specified rights to act on src file/folder or
      registry key.
    type: str
    required: yes
  state:
    description:
    - Specify whether to add C(present) or remove C(absent) the specified access rule.
    type: str
    choices: [ absent, present ]
    default: present
  type:
    description:
    - Specify whether to allow or deny the rights specified.
    type: str
    required: yes
    choices: [ allow, deny ]
  rights:
    description:
    - The rights/permissions that are to be allowed/denied for the specified
      user or group for the item at C(path).
    - If C(path) is a file or directory, rights can be any right under MSDN
      FileSystemRights U(https://msdn.microsoft.com/en-us/library/system.security.accesscontrol.filesystemrights.aspx).
    - If C(path) is a registry key, rights can be any right under MSDN
      RegistryRights U(https://msdn.microsoft.com/en-us/library/system.security.accesscontrol.registryrights.aspx).
    - If I(path) is a certificate key, rights can be C(Read) and/or C(FullControl). (Added in 2.2.0)
    type: str
    required: yes
  inherit:
    description:
    - Inherit flags on the ACL rules.
    - Can be specified as a comma separated list, e.g. C(ContainerInherit),
      C(ObjectInherit).
    - For more information on the choices see MSDN InheritanceFlags enumeration
      at U(https://msdn.microsoft.com/en-us/library/system.security.accesscontrol.inheritanceflags.aspx).
    - Defaults to C(ContainerInherit, ObjectInherit) for Directories.
    type: str
    choices: [ ContainerInherit, ObjectInherit ]
  propagation:
    description:
    - Propagation flag on the ACL rules.
    - For more information on the choices see MSDN PropagationFlags enumeration
      at U(https://msdn.microsoft.com/en-us/library/system.security.accesscontrol.propagationflags.aspx).
    type: str
    choices: [ InheritOnly, None, NoPropagateInherit ]
    default: "None"
  follow:
    description:
    - Follow the symlinks and junctions to apply the ACLs to the target instead of the link.
    type: bool
    default: false
    version_added: 1.12.0
notes:
- If adding ACL's for AppPool identities, the Windows Feature "Web-Scripting-Tools" must be enabled.
seealso:
- module: ansible.windows.win_acl_inheritance
- module: ansible.windows.win_file
- module: ansible.windows.win_owner
- module: ansible.windows.win_stat
author:
- Phil Schwartz (@schwartzmx)
- Trond Hindenes (@trondhindenes)
- Hans-Joachim Kliemeck (@h0nIg)
'''

EXAMPLES = r'''
- name: Restrict write and execute access to User Fed-Phil
  ansible.windows.win_acl:
    user: Fed-Phil
    path: C:\Important\Executable.exe
    type: deny
    rights: ExecuteFile,Write

- name: Add IIS_IUSRS allow rights
  ansible.windows.win_acl:
    path: C:\inetpub\wwwroot\MySite
    user: IIS_IUSRS
    rights: FullControl
    type: allow
    state: present
    inherit: ContainerInherit, ObjectInherit
    propagation: 'None'

- name: Set registry key right
  ansible.windows.win_acl:
    path: HKCU:\Bovine\Key
    user: BUILTIN\Users
    rights: EnumerateSubKeys
    type: allow
    state: present
    inherit: ContainerInherit, ObjectInherit
    propagation: 'None'

- name: Remove FullControl AccessRule for IIS_IUSRS
  ansible.windows.win_acl:
    path: C:\inetpub\wwwroot\MySite
    user: IIS_IUSRS
    rights: FullControl
    type: allow
    state: absent
    inherit: ContainerInherit, ObjectInherit
    propagation: 'None'

- name: Deny Intern
  ansible.windows.win_acl:
    path: C:\Administrator\Documents
    user: Intern
    rights: Read,Write,Modify,FullControl,Delete
    type: deny
    state: present

- name: Set certificate private key FullControl to IIS_IUSRS
  ansible.windows.win_acl:
    path: Cert:\LocalMachine\My\168ba8c488463f88c6648466a22484b6189e165f
    user: IIS_IUSRS
    type: allow
    state: present
    rights: FullControl
'''



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Akini Ross (@akinross) <akinross@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_l3_domain
short_description: Manage L3 Domains on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage L3 Domains on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.3) and later.
author:
- Akini Ross (@akinross)
options:
  template:
    description:
    - The name of the template.
    - The template must be a fabric policy template.
    type: str
    required: true
  l3_domain:
    description:
    - The name of the L3 Domain.
    type: str
    aliases: [ name ]
  l3_domain_uuid:
    description:
    - The uuid of the L3 Domain.
    - This parameter is required when the O(l3_domain) needs to be updated.
    type: str
    aliases: [ uuid ]
  description:
    description:
    - The description of the L3 Domain.
    type: str
  pool:
    description:
    - The name of the VLAN Pool.
    - Providing an empty string will remove the pool from the L3 Domain.
    type: str
    aliases: [ vlan_pool ]
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create a new L3 domain
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    l3_domain: ansible_test_l3_domain
    pool: ansible_test_vlan_pool
    state: present
  register: create

- name: Query a L3 domain with name
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    l3_domain: ansible_test_l3_domain
    state: query
  register: query_one

- name: Query a L3 domain with UUID
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    l3_domain_uuid: '{{ create.current.uuid }}'
    state: query
  register: query_with_uuid

- name: Query all L3 domains in the template
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    state: query
  register: query_all

- name: Delete a vlan pool from a L3 domain
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    l3_domain: ansible_test_l3_domain
    pool: ""
    state: present

- name: Delete a L3 domain
  cisco.mso.ndo_l3_domain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_tenant_template
    l3_domain: ansible_test_l3_domain
    state: absent
"""

RETURN = r"""
"""


import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate, KVPair


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str", required=True),
        l3_domain=dict(type="str", aliases=["name"]),
        l3_domain_uuid=dict(type="str", aliases=["uuid"]),
        description=dict(type="str"),
        pool=dict(type="str", aliases=["vlan_pool"]),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["l3_domain"]],
            ["state", "present", ["l3_domain"]],
        ],
    )

    mso = MSOModule(module)

    template = module.params.get("template")
    l3_domain = module.params.get("l3_domain")
    l3_domain_uuid = module.params.get("l3_domain_uuid")
    pool = module.params.get("pool")
    description = module.params.get("description")
    state = module.params.get("state")

    ops = []
    match = None

    mso_template = MSOTemplate(mso, "fabric_policy", template)
    mso_template.validate_template("fabricPolicy")

    path = "/fabricPolicyTemplate/template/l3Domains"
    match = get_l3_domain(mso_template, l3_domain_uuid, l3_domain)
    if l3_domain_uuid or l3_domain:
        if match:
            if match.details.get("pool"):
                match.details["poolRef"] = match.details.get("pool")
                match.details["pool"] = mso_template.get_vlan_pool_name(match.details.get("pool"))
            mso.existing = mso.previous = copy.deepcopy(match.details)  # Query a specific object
    elif match:
        for l3_domain in match:
            if l3_domain.get("pool"):
                l3_domain["poolRef"] = l3_domain.get("pool")
                l3_domain["pool"] = mso_template.get_vlan_pool_name(l3_domain.get("pool"))
        mso.existing = match  # Query all objects

    if state == "present":
        mso.existing = {}

        if match:
            if l3_domain and match.details.get("name") != l3_domain:
                ops.append(dict(op="replace", path="{0}/{1}/name".format(path, match.index), value=l3_domain))
                match.details["name"] = l3_domain

            if description is not None and match.details.get("description") != description:
                ops.append(dict(op="replace", path="{0}/{1}/description".format(path, match.index), value=description))
                match.details["description"] = description

            if pool and match.details.get("pool") != pool:
                ops.append(dict(op="replace", path="{0}/{1}/pool".format(path, match.index), value=mso_template.get_vlan_pool_uuid(pool)))
                match.details["pool"] = pool
            elif pool == "" and match.details.get("pool"):
                ops.append(dict(op="remove", path="{0}/{1}/pool".format(path, match.index)))
                match.details.pop("pool")

            mso.sanitize(match.details)

        else:
            payload = {"name": l3_domain, "templateId": mso_template.template.get("templateId"), "schemaId": mso_template.template.get("schemaId")}
            if description:
                payload["description"] = description

            if pool:
                pool_uuid = mso_template.get_vlan_pool_uuid(pool)
                payload["pool"] = pool_uuid

            ops.append(dict(op="add", path="{0}/-".format(path), value=copy.deepcopy(payload)))

            if pool:
                payload["pool"] = pool
                payload["poolRef"] = pool_uuid

            mso.sanitize(payload)

        mso.existing = mso.proposed

    elif state == "absent":
        if match:
            ops.append(dict(op="remove", path="{0}/{1}".format(path, match.index)))
        mso.existing = {}

    if not module.check_mode and ops:
        mso_template.template = mso.request(mso_template.template_path, method="PATCH", data=ops)
        match = get_l3_domain(mso_template, l3_domain_uuid, l3_domain)
        if match:
            if match.details.get("pool"):
                match.details["poolRef"] = match.details.get("pool")
                match.details["pool"] = mso_template.get_vlan_pool_name(match.details.get("pool"))
            mso.existing = match.details  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.proposed["poolRef"] = mso_template.get_vlan_pool_uuid(pool)
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


def get_l3_domain(mso_template, uuid=None, name=None, fail_module=False):
    match = mso_template.template.get("fabricPolicyTemplate", {}).get("template", {}).get("l3Domains", [])
    if uuid or name:  # Query a specific object
        return mso_template.get_object_by_key_value_pairs("L3 Domain", match, [KVPair("uuid", uuid) if uuid else KVPair("name", name)], fail_module)
    return match  # Query all objects


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_san_connectivity_policy_info
short_description: Gather information about SAN Connectivity Policies and their vHBAs in Cisco Intersight
description:
  - Retrieve information about SAN Connectivity Policies and their associated vHBAs from L(Cisco Intersight,https://intersight.com).
  - Query policies by organization, policy name, or filter vHBAs by name patterns.
  - Returns structured data combining policy metadata with detailed vHBA configurations.
  - Supports filtering by organization, policy name, and vHBA name.
  - If no filters are provided, all SAN Connectivity Policies and their vHBAs will be returned.
  - Returns structured data with both policy information and associated vHBA details.
extends_documentation_fragment: intersight
options:
  organization:
    description:
      - The name of the organization to filter SAN Connectivity Policies by.
      - Use 'default' for the default organization.
      - When specified, only policies from this organization will be returned.
    type: str
  name:
    description:
      - The exact name of the SAN Connectivity Policy to retrieve information from.
      - When specified, only the matching policy and its vHBAs will be returned.
      - Can be combined with vhba_name filter to get specific vHBAs from a specific policy.
    type: str
  vhba_name:
    description:
      - Filter vHBAs by name within the policies.
      - Can be used with or without policy name filtering.
      - Supports exact matching (e.g., "vhba-a" will match only "vhba-a").
      - When combined with policy name, filters vHBAs within that specific policy.
    type: str
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
# Basic Usage Examples
- name: Fetch all SAN Connectivity Policies and their vHBAs from all organizations
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
  register: all_san_policies

# Organization-specific Examples
- name: Fetch all SAN Connectivity Policies from the default organization
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
  register: default_org_policies

- name: Fetch all SAN Connectivity Policies from a custom organization
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "Engineering"
  register: engineering_policies

# Specific Policy Examples
- name: Fetch a specific SAN Connectivity Policy by name with all its vHBAs
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "fi-attached-san-policy"
  register: specific_policy

# vHBA Filtering Examples
- name: Find all policies containing a specific vHBA
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    vhba_name: "vhba-a"
  register: policies_with_vhba_a

- name: Find specific vHBA in a specific policy
  cisco.intersight.intersight_san_connectivity_policy_info:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    name: "fi-attached-san-policy"
    vhba_name: "vhba-b"
  register: specific_vhba_in_policy
'''

RETURN = r'''
api_response:
  description:
    - The API response containing policy and vHBA information.
    - Returns a dictionary when querying a single policy or no policies found.
    - Returns a list when multiple policies are found.
  returned: always
  type: dict
  sample:
    # Single policy response (when name parameter is used or only one policy found)
    Name: "fi-attached-san-policy"
    ObjectType: "vnic.SanConnectivityPolicy"
    Moid: "12345678901234567890abcd"
    Description: "SAN connectivity policy for FI-attached servers"
    TargetPlatform: "FIAttached"
    WwnnAddressType: "POOL"
    PlacementMode: "custom"
    Organization:
      Name: "default"
      ObjectType: "organization.Organization"
      Moid: "abcdef1234567890abcdef12"
    Tags:
      - Key: "Environment"
        Value: "Production"
    WwnnPool:
      Moid: "fedcba0987654321fedcba09"
      ObjectType: "fcpool.Pool"
    vhbas:
      - Name: "vhba-a"
        ObjectType: "vnic.FcIf"
        Moid: "vhba12345678901234567890"
        Order: 0
        PersistentBindings: true
        Type: "fc-initiator"
        Placement:
          SwitchId: "A"
          PciLink: 0
        WwpnAddressType: "POOL"
        SanConnectivityPolicy:
          Moid: "12345678901234567890abcd"
          ObjectType: "vnic.SanConnectivityPolicy"
      - Name: "vhba-b"
        ObjectType: "vnic.FcIf"
        Moid: "vhba09876543210987654321"
        Order: 1
        PersistentBindings: true
        Type: "fc-initiator"
        Placement:
          SwitchId: "B"
          PciLink: 0
        WwpnAddressType: "POOL"
        SanConnectivityPolicy:
          Moid: "12345678901234567890abcd"
          ObjectType: "vnic.SanConnectivityPolicy"
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def get_vhbas_for_policy(intersight, policy_moid, vhba_name_filter=None):
    """
    Get all vHBAs associated with a specific policy.
    """

    # Build filter for vHBAs associated with this policy
    filter_conditions = [f"SanConnectivityPolicy.Moid eq '{policy_moid}'"]

    # Add vHBA name filter if provided (exact match only)
    if vhba_name_filter:
        filter_conditions.append(f"Name eq '{vhba_name_filter}'")

    query_params = {
        '$filter': ' and '.join(filter_conditions)
    }

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get vHBAs for this policy
    intersight.get_resource(
        resource_path='/vnic/FcIfs',
        query_params=query_params,
        return_list=True
    )

    # Capture the response immediately before it gets overwritten
    vhbas_response = intersight.result.get('api_response', [])

    # Ensure we return a list
    if isinstance(vhbas_response, list):
        return vhbas_response
    elif isinstance(vhbas_response, dict):
        return [vhbas_response]
    else:
        return []


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        organization=dict(type='str'),
        name=dict(type='str'),
        vhba_name=dict(type='str')
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)

    # Resource path used to fetch policy info
    resource_path = '/vnic/SanConnectivityPolicies'

    # Get query parameters for policies
    query_params = intersight.set_query_params()

    # Reset api_response before the API call to avoid previous responses
    intersight.result['api_response'] = {}

    # Get SAN Connectivity policies
    intersight.get_resource(
        resource_path=resource_path,
        query_params=query_params,
        return_list=True
    )

    policies = intersight.result.get('api_response', [])
    vhba_name_filter = module.params.get('vhba_name')

    # Build structured response
    structured_results = []

    # Ensure policies is always a list for iteration, even if a single dict is returned
    if isinstance(policies, dict):
        policies = [policies]
    elif not isinstance(policies, list):
        policies = []

    for policy in policies:
        policy_moid = policy.get('Moid')
        if not policy_moid:
            continue

        # Get vHBAs for this policy
        vhbas = get_vhbas_for_policy(intersight, policy_moid, vhba_name_filter)

        # Embed vHBAs in the policy response
        if vhbas:
            policy['vhbas'] = vhbas

        structured_results.append(policy)

    # Create final response structure
    final_api_response = None

    # Set final response based on number of policies found
    if len(structured_results) == 1:
        # Single policy - return as dict
        final_api_response = structured_results[0]
    elif len(structured_results) > 1:
        # Multiple policies - return as list
        final_api_response = structured_results
    else:
        # No policies found - return empty structure
        final_api_response = {}

    # Use intersight.result and update api_response directly
    intersight.result['api_response'] = final_api_response
    final_result = intersight.result

    module.exit_json(**final_result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_applicationfirewallpolicy
version_added: "3.1.0"
short_description: Manage the Application firewall policy instance
description:
    - Creating, Updating or Deleting the application firewall policy instance.
options:
    name:
        description:
            - The name of the application firewall policy's name.
        type: str
        required: true
    resource_group:
        description:
            - The name of the resource group.,
        type: str
        required: true
    location:
        description:
            - Valid Azure location. Defaults to the location of the resource group.
        type: str
    policy_settings:
        description:
            - The PolicySettings for policy.
        type: dict
        suboptions:
            state:
                description:
                    - The state of the policy.
                type: str
                choices:
                    - Disabled
                    - Enabled
            mode:
                description:
                    - The mode of the policy.
                type: str
                choices:
                    - Prevention
                    - Detection
            request_body_check:
                description:
                    - Whether to allow WAF to check request Body.
                type: bool
            request_body_inspect_limit_in_kb:
                description:
                    - Max inspection limit in KB for request body inspection for WAF.
                type: int
            request_body_enforcement:
                description:
                    - Whether allow WAF to enforce request body limits.
                type: bool
            max_request_body_size_in_kb:
                description:
                    - Maximum request body size in Kb for WAF.
                type: int
            file_upload_enforcement:
                description:
                    - Whether allow WAF to enforce file upload limits.
                type: bool
            file_upload_limit_in_mb:
                description:
                    - Maximum file upload size in Mb for WAF.
                type: int
            custom_block_response_status_code:
                description:
                    - If the action type is block, customer can override the response status code.
                type: int
            custom_block_response_body:
                description:
                    - If the action type is block, customer can override the response body.
                    - The body must be specified in base64 encoding.
                type: str
            js_challenge_cookie_expiration_in_mins:
                description:
                    - Web Application Firewall JavaScript Challenge Cookie Expiration time in minutes.
                type: int
            log_scrubbing:
                description:
                    - To scrub sensitive log fields.
                type: dict
                suboptions:
                    state:
                        description:
                            - State of the log scrubbing config.
                        type: str
                        choices:
                            - Enabled
                            - Disabled
                    scrubbing_rules:
                        description:
                            - The rules that are applied to the logs for scrubbing.
                        type: list
                        elements: dict
                        suboptions:
                            match_variable:
                                description:
                                    - The variable to be scrubbed from the logs.
                                type: str
                                required: true
                                choices:
                                    - RequestHeaderNames
                                    - RequestCookieNames
                                    - RequestArgNames
                                    - RequestPostArgNames
                                    - RequestJSONArgNames
                                    - RequestIPAddress
                            selector_match_operator:
                                description:
                                    - "When matchVariable is a collection, operate on the selector to specify which
                                       elements in the collection this rule applies to."
                                required: true
                                type: str
                                choices:
                                    - Equals
                                    - EqualsAny
                            selector:
                                description:
                                    - When matchVariable is a collection, operator used to specify which elements in the collection this rule applies to.
                                type: str
                            state:
                                description:
                                    - Defines the state of log scrubbing rule.
                                type: str
                                choices:
                                    - Enabled
                                    - Disabled
    custom_rules:
        description:
            - The custom rules inside the policy.
        type: list
        elements: dict
        suboptions:
            priority:
                description:
                    - Priority of the rule.
                    - Rules with a lower value will be evaluated before rules with a higher value.
                required: true
                type: int
            rule_type:
                description:
                    - The rule type.
                required: true
                type: str
                choices:
                    - MatchRule
                    - RateLimitRule
                    - Invalid
            match_conditions:
                description:
                    - List of match conditions.
                type: list
                elements: dict
                required: true
                suboptions:
                    match_variables:
                        description:
                            - List of match variables.
                        required: true
                        type: list
                        elements: dict
                        suboptions:
                            variable_name:
                                description:
                                    - Match Variable.
                                required: true
                                type: str
                                choices:
                                    - RemoteAddr
                                    - RequestMethod
                                    - QueryString
                                    - PostArgs
                                    - RequestUri
                                    - RequestHeaders
                                    - RequestBody
                                    - RequestCookies
                            selector:
                                description:
                                    - The selector of match variable.
                                type: str
                    operator:
                        description:
                            - The operator to be matched.
                        required: true
                        type: str
                        choices:
                            - IPMatch
                            - Equal
                            - Contains
                            - LessThan
                            - GreaterThan
                            - LessThanOrEqual
                            - GreaterThanOrEqual
                            - BeginsWith
                            - EndsWith
                            - Regex
                            - GeoMatch
                            - Any
                    negation_conditon:
                        description:
                            - Whether this is negate condition or not.
                        type: bool
                    match_values:
                        description:
                            - Match value.
                        required: true
                        type: list
                        elements: str
                    transforms:
                        description:
                            - List of transforms.
                        type: list
                        elements: str
                        choices:
                            - Uppercase
                            - Lowercase
                            - Trim
                            - UrlDecode
                            - UrlEncode
                            - RemoveNulls
                            - HtmlEntityDecode
            action:
                description:
                    - Type of Actions.
                type: str
                choices:
                    - Allow
                    - Block
                    - Log
                    - JSChallenge
            name:
                description:
                    - The name of the resource that is unique within a policy.
                    - This name can be used to access the resource.
                type: str
            state:
                description:
                    - Describes if the custom rule is in enabled or disabled state. Defaults to Enabled if not specified.
                type: str
                choices:
                    - Disabled
                    - Enabled
            rate_limit_duration:
                description:
                    - Duration over which Rate Limit policy will be applied.
                    - Applies only when ruleType is RateLimitRule.
                type: str
                choices:
                    - OneMin
                    - FiveMins
            rate_limit_threshold:
                description:
                    - Rate Limit threshold to apply in case ruleType is RateLimitRule.
                    - Must be greater than or equal to 1.
                type: int
            group_by_user_session:
                description:
                    - List of user session identifier group by clauses.
                type: list
                elements: dict
                suboptions:
                    group_by_variables:
                        description:
                            - List of group by clause variables.
                        type: list
                        elements: dict
                        suboptions:
                            variable_name:
                                description:
                                    - User Session clause variable.
                                type: str
                                choices:
                                    - ClientAddr
                                    - GeoLocation
                                    - None
    managed_rules:
        description:
            - Describes the managedRules structure.
        type: dict
        suboptions:
            exclusions:
                description:
                    - The Exclusions that are applied on the policy.
                type: list
                elements: dict
                suboptions:
                    match_variable:
                        description:
                            - The variable to be excluded.
                        required: true
                        type: str
                        choices:
                            - RequestHeaderNames
                            - RequestCookieNames
                            - RequestArgNames
                            - RequestHeaderKeys
                            - RequestHeaderValues
                            - RequestCookieKeys
                            - RequestCookieValues
                            - RequestArgKeys
                            - RequestArgValues
                    selector_match_operator:
                        description:
                            - When matchVariable is a collection, operate on the selector to specify which elements in the collection this exclusion applies to.
                        required: true
                        type: str
                        choices:
                            - Equals
                            - Contains
                            - StartsWith
                            - EndsWith
                            - EqualsAny
                    selector:
                        description:
                            - When matchVariable is a collection, operator used to specify which elements in the collection this exclusion applies to.
                        required: true
                        type: str
                    exclusion_managed_rule_sets:
                        description:
                            - The managed rule sets that are associated with the exclusion.
                        type: list
                        elements: dict
                        suboptions:
                            rule_set_type:
                                description:
                                    - Defines the rule set type to use.
                                type: str
                                required: true
                            rule_set_version:
                                description:
                                    - Defines the version of the rule set to use.
                                type: str
                                required: true
                            rule_groups:
                                description:
                                    - Defines the rule groups to apply to the rule set.
                                type: list
                                elements: dict
                                suboptions:
                                    rule_group_name:
                                        description:
                                            - The managed rule group for exclusion.
                                        type: str
                                        required: true
                                    rules:
                                        description:
                                            - List of rules that will be excluded. If none specified, all rules in the group will be excluded.
                                        type: list
                                        elements: dict
                                        suboptions:
                                            rule_id:
                                                description:
                                                    - Identifier for the managed rule.
                                                type: str
            managed_rule_sets:
                description:
                    - The managed rule sets that are associated with the policy.
                required: true
                type: list
                elements: dict
                suboptions:
                    rule_set_type:
                        description:
                            - Defines the rule set type to use.
                        type: str
                        required: true
                    rule_set_version:
                        description:
                            - Defines the version of the rule set to use.
                        required: true
                        type: str
                    rule_group_overrides:
                        description:
                            - Defines the rule group overrides to apply to the rule set.
                        type: list
                        elements: dict
                        suboptions:
                            rule_group_name:
                                description:
                                    - The managed rule group to override.
                                required: true
                                type: str
                            rules:
                                description:
                                    - List of rules that will be disabled.
                                    - If none specified, all rules in the group will be disabled.
                                type: list
                                elements: dict
                                suboptions:
                                    rule_id:
                                        description:
                                            - Identifier for the managed rule.
                                        required: true
                                        type: str
                                    state:
                                        description:
                                            - The state of the managed rule. Defaults to C(Disabled) if not specified.
                                        type: str
                                        choices:
                                            - Enabled
                                            - Disabled
                                    action:
                                        description:
                                            - Describes the override action to be applied when rule matches.
                                        type: str
                                        choices:
                                            - AnomalyScoring
                                            - Allow
                                            - Block
                                            - Log
                                            - JSChallenge
                                    sensitivity:
                                        description:
                                            - Describes the override sensitivity to be applied when rule matches.
                                        type: str
                                        choices:
                                            - None
                                            - Low
                                            - Medium
                                            - High
    state:
        description:
            - Assert the state of the firewall policy.
            - Use C(present) to create or update a and C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Create a new application firewall policy
  azure_rm_applicationfirewallpolicy:
    resource_group: "{{ resource_group }}"
    name: "new{{ rpfx }}02"
    location: australiasoutheast
    policy_settings:
      file_upload_enforcement: true
      file_upload_limit_in_mb: 88
      js_challenge_cookie_expiration_in_mins: 30
      max_request_body_size_in_kb: 128
      mode: Detection
      request_body_check: true
      request_body_enforcement: true
      request_body_inspect_limit_in_kb: 128
      state: Enabled
      custom_block_response_status_code: 200
      custom_block_response_body: Fredtest
      log_scrubbing:
        state: Enabled
        scrubbing_rules:
          - match_variable: RequestHeaderNames
            selector_match_operator: Equals
            selector: '*'
            state: Enabled
    custom_rules:
      - action: Block
        match_conditions:
          - match_values:
              - 10.1.0.0/24
              - 10.2.0.0/24
            match_variables:
              - variable_name: RemoteAddr
            negation_conditon: true
            operator: IPMatch
            transforms:
              - Uppercase
              - Lowercase
        name: ruledefine01
        priority: 21
        rule_type: MatchRule
        state: Enabled
      - action: Block
        group_by_user_session:
          - group_by_variables:
              - variable_name: ClientAddr
        match_conditions:
          - match_values:
              - 10.1.0.0/24
              - 10.2.0.0/24
            match_variables:
              - variable_name: RemoteAddr
            negation_conditon: false
            operator: IPMatch
        name: ruledefine02
        priority: 22
        rule_type: RateLimitRule
        rate_limit_threshold: 100
        rate_limit_duration: OneMin
        state: Enabled
    managed_rules:
      exclusions:
        - match_variable: RequestHeaderNames
          selector_match_operator: Equals
          selector: IPMatch
          exclusion_managed_rule_sets:
            - rule_set_type: Microsoft_DefaultRuleSet
              rule_set_version: 2.1
      managed_rule_sets:
        - rule_set_type: Microsoft_BotManagerRuleSet
          rule_set_version: 1.0
        - rule_set_type: Microsoft_DefaultRuleSet
          rule_set_version: 2.1

- name: Delete the application firewall policy
  azure_rm_applicationfirewallpolicy:
    resource_group: "{{ resource_group }}"
    name: firewallpolicy
    state: absent
'''

RETURN = '''
firewall_policy:
    description:
        - A list of the application firewall policy facts
    returned: always
    type: complex
    contains:
        id:
            description:
                - The application firewall policy's ID.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/v-xisuRG/providers/Microsoft.Network/ApplicationGatewayWebApplicationFirewallPolicies/firewallpolicy"
        name:
            description:
                - Name of application firewall policy.
            returned: always
            type: str
            sample: firewallpolicy
        resource_group:
            description:
                - Name of resource group.
            returned: always
            type: str
            sample: myResourceGroup
        location:
            description:
                - Location of application firewall policy.
            returned: always
            type: str
            sample: eastus
        provisioning_state:
            description:
                - Provisioning state of application firewall policy.
            returned: always
            type: str
            sample: Succeeded
        type:
            description:
                - The type of the application firewall policy.
            returned: always
            type: str
            sample: Microsoft.Network/ApplicationGatewayWebApplicationFirewallPolicies
        tags:
            description:
                - The application firewall policy tags.
            type: dict
            returned: always
            sample: {"key1": "value1"}
        custom_rules:
            description:
                - The custom rules inside the policy.
            type: complex
            returned: when used
            contains:
                action:
                    description:
                        - The name of the resource that is unique within a policy.
                        - This name can be used to access the resource.
                    type: str
                    returned: always
                    sample: Block
                match_conditions:
                    description:
                        - List of match conditions.
                    type: list
                    returned: always
                    sample: [{'match_values': ['10.1.0.4'], 'match_variables': [{'variable_name': 'RemoteAddr'}],
                              'negation_condition': false, 'operator': 'IPMatch', 'transforms': []}]
                name:
                    description:
                        - The name of the resource that is unique within a policy.
                        - This name can be used to access the resource.
                    type: str
                    returned: always
                    sample: testrule01
                priority:
                    description:
                        - Priority of the rule.
                        - Rules with a lower value will be evaluated before rules with a higher value.
                    type: int
                    returned: always
                    sample: 33
                rule_type:
                    description:
                        - The rule type.
                    type: str
                    returned: always
                    sample: MatchRule
                state:
                    description:
                        - Describes if the custom rule is in enabled or disabled state.
                    type: str
                    returned: always
                    sample: Enabled
        managed_rules:
            description:
                - Describes the managedRules structure.
            type: complex
            returned: when used
            contains:
                exclusions:
                    description:
                        - The exceptions that are applied on the policy.
                    type: list
                    returned: always
                    sample: []
                managed_rule_sets:
                    description:
                        - The managed rule sets that are associated with the policy.
                    type: list
                    returned: always
                    sample: [{"rule_group_overrides": [],
                              "rule_set_type": "Microsoft_DefaultRuleSet",
                              "rule_set_version": "2.1"
                             }]
        policy_settings:
            description:
                - The PolicySettings for policy.
            type: complex
            returned: when used
            contains:
                file_upload_enforcement:
                    description:
                        - Whether allow WAF to enforce file upload limits.
                    type: bool
                    returned: always
                    sample: true
                file_upload_limit_in_mb:
                    description:
                        - Maximum file upload size in Mb for WAF.
                    type: int
                    returned: always
                    sample: 100
                js_challenge_cookie_expiration_in_mins:
                    description:
                        - Web Application Firewall JavaScript Challenge Cookie Expiration time in minutes.
                    type: int
                    returned: always
                    sample: 30
                max_request_body_size_in_kb:
                    description:
                        - Maximum request body size in Kb for WAF.
                    type: int
                    returned: always
                    sample: 128
                mode:
                    description:
                        - The mode of the policy.
                    type: str
                    returned: always
                    sample: Detection
                request_body_check:
                    description:
                        - Whether to allow WAF to check request Body.
                    type: bool
                    returned: always
                    sample: false
                request_body_enforcement:
                    description:
                        - Whether allow WAF to enforce request body limits.
                    type: bool
                    returned: always
                    sample: false
                state:
                    description:
                        - The state of the policy.
                    type: str
                    returned: always
                    sample: Enabled
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt

try:
    from azure.core.exceptions import ResourceNotFoundError
    from azure.mgmt.core.tools import parse_resource_id
except ImportError:
    # This is handled in azure_rm_common
    pass


policy_setting_spec = dict(
    state=dict(type='str', choices=['Disabled', 'Enabled']),
    mode=dict(type='str', choices=['Prevention', 'Detection']),
    request_body_check=dict(type='bool'),
    request_body_inspect_limit_in_kb=dict(type='int'),
    request_body_enforcement=dict(type='bool'),
    max_request_body_size_in_kb=dict(type='int'),
    file_upload_enforcement=dict(type='bool'),
    file_upload_limit_in_mb=dict(type='int'),
    custom_block_response_status_code=dict(type='int'),
    custom_block_response_body=dict(type='str'),
    js_challenge_cookie_expiration_in_mins=dict(type='int'),
    log_scrubbing=dict(
        type='dict',
        options=dict(
            state=dict(type='str', choices=['Enabled', 'Disabled']),
            scrubbing_rules=dict(
                type='list',
                elements='dict',
                options=dict(
                    match_variable=dict(
                        type='str',
                        required=True,
                        choices=["RequestHeaderNames", "RequestCookieNames", "RequestArgNames",
                                 "RequestPostArgNames", "RequestJSONArgNames", "RequestIPAddress"]
                    ),
                    selector_match_operator=dict(type='str', required=True, choices=["Equals", "EqualsAny"]),
                    selector=dict(type='str'),
                    state=dict(type='str', choices=['Enabled', 'Disabled']),
                )
            )
        )
    )
)


custom_rule_spec = dict(
    priority=dict(type='int', required=True),
    rule_type=dict(type='str', required=True, choices=['MatchRule', 'RateLimitRule', 'Invalid']),
    match_conditions=dict(
        type='list',
        elements='dict',
        required=True,
        options=dict(
            match_variables=dict(
                type='list',
                required=True,
                elements='dict',
                options=dict(
                    variable_name=dict(
                        type='str',
                        required=True,
                        choices=["RemoteAddr", "RequestMethod", "QueryString", "PostArgs", "RequestUri", "RequestHeaders", "RequestBody", "RequestCookies"]
                    ),
                    selector=dict(type='str')
                )
            ),
            operator=dict(
                type='str',
                required=True,
                choices=["IPMatch", "Equal", "Contains", "LessThan", "GreaterThan", "LessThanOrEqual",
                         "GreaterThanOrEqual", "BeginsWith", "EndsWith", "Regex", "GeoMatch", "Any"]
            ),
            match_values=dict(
                type='list',
                elements='str',
                required=True
            ),
            transforms=dict(
                type='list',
                elements='str',
                choices=['Uppercase', 'Lowercase', 'Trim', 'UrlDecode', 'UrlEncode', 'RemoveNulls', 'HtmlEntityDecode']
            ),
            negation_conditon=dict(
                type='bool'
            ),
        )
    ),
    action=dict(type='str', choices=['Allow', 'Block', 'Log', 'JSChallenge']),
    name=dict(type='str'),
    state=dict(type='str', choices=['Disabled', 'Enabled']),
    rate_limit_duration=dict(type='str', choices=['OneMin', 'FiveMins']),
    rate_limit_threshold=dict(type='int'),
    group_by_user_session=dict(
        type='list',
        elements='dict',
        options=dict(
            group_by_variables=dict(
                type='list',
                elements='dict',
                options=dict(
                    variable_name=dict(
                        type='str',
                        choices=["ClientAddr", "GeoLocation", "None"]
                    )
                )
            )
        )
    )
)


managed_rule_spec = dict(
    managed_rule_sets=dict(
        type='list',
        elements='dict',
        required=True,
        options=dict(
            rule_set_type=dict(type='str', required=True),
            rule_set_version=dict(type='str', required=True),
            rule_group_overrides=dict(
                type='list',
                elements='dict',
                options=dict(
                    rule_group_name=dict(type='str', required=True),
                    rules=dict(
                        type='list',
                        elements='dict',
                        options=dict(
                            rule_id=dict(type='str', required=True),
                            state=dict(type='str', choices=['Enabled', 'Disabled']),
                            action=dict(type='str', choices=["AnomalyScoring", "Allow", "Block", "Log", "JSChallenge"]),
                            sensitivity=dict(type='str', choices=["None", "Low", "Medium", "High"])
                        )
                    )
                )
            )
        )
    ),
    exclusions=dict(
        type='list',
        elements='dict',
        options=dict(
            match_variable=dict(
                type='str',
                required=True,
                choices=["RequestHeaderNames", "RequestCookieNames", "RequestArgNames", "RequestHeaderKeys",
                         "RequestHeaderValues", "RequestCookieKeys", "RequestCookieValues", "RequestArgKeys", "RequestArgValues"]
            ),
            selector_match_operator=dict(
                type='str',
                required=True,
                choices=["Equals", "Contains", "StartsWith", "EndsWith", "EqualsAny"]
            ),
            selector=dict(
                type='str',
                required=True
            ),
            exclusion_managed_rule_sets=dict(
                type='list',
                elements='dict',
                options=dict(
                    rule_set_type=dict(type='str', required=True),
                    rule_set_version=dict(type='str', required=True),
                    rule_groups=dict(
                        type='list',
                        elements='dict',
                        options=dict(
                            rule_group_name=dict(type='str', required=True),
                            rules=dict(
                                type='list',
                                elements='dict',
                                options=dict(
                                    rule_id=dict(type='str')
                                )
                            )
                        )
                    )
                )
            )
        )
    )
)


class AzureRMApplicationFirewallPolicy(AzureRMModuleBaseExt):

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            location=dict(type='str'),
            policy_settings=dict(type='dict', options=policy_setting_spec),
            custom_rules=dict(type='list', elements='dict', options=custom_rule_spec),
            managed_rules=dict(type='dict', options=managed_rule_spec),
            state=dict(type='str', choices=['present', 'absent'], default='present')
        )

        self.results = dict(
            changed=False,
        )

        self.name = None
        self.resource_group = None
        self.state = None
        self.body = dict()

        super(AzureRMApplicationFirewallPolicy, self).__init__(self.module_arg_spec,
                                                               supports_check_mode=True,
                                                               supports_tags=True,
                                                               facts_module=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif kwargs[key] is not None:
                self.body[key] = kwargs[key]

        old_response = self.get()
        changed = False

        resource_group = self.get_resource_group(self.resource_group)
        if not self.body.get('location'):
            # Set default location
            self.body['location'] = resource_group.location

        if old_response is not None:
            if self.state == 'present':

                if not self.default_compare({}, self.body, old_response, '', dict(compare=[])):
                    changed = True
                    if not self.check_mode:
                        self.create_or_update(self.body)
            else:
                changed = True
                if not self.check_mode:
                    self.delete()
        else:
            if self.state == 'present':
                changed = True
                if not self.check_mode:
                    self.create_or_update(self.body)

        self.results["firewall_policy"] = self.get()
        self.results['changed'] = changed

        return self.results

    def get(self):
        response = None
        try:
            response = self.network_client.web_application_firewall_policies.get(resource_group_name=self.resource_group, policy_name=self.name)
        except ResourceNotFoundError:
            pass

        if response is not None:
            return self.format_response(response)

        return response

    def create_or_update(self, body):
        response = None
        try:
            response = self.network_client.web_application_firewall_policies.create_or_update(resource_group_name=self.resource_group,
                                                                                              policy_name=self.name,
                                                                                              parameters=body)
        except Exception as exc:
            self.fail("Error creating or update the application firewall policy in resource groups {0}: {1}".format(self.resource_group, str(exc)))

        if response is not None:
            return self.format_response(response)
        else:
            return None

    def delete(self):
        try:
            self.network_client.web_application_firewall_policies.begin_delete(resource_group_name=self.resource_group,
                                                                               policy_name=self.name)
        except Exception as exc:
            self.fail("Error deleting the application firewall policy: {0}".format(str(exc)))

        return None

    def format_response(self, item):
        d = item.as_dict()
        id_dict = parse_resource_id(item.id)
        d['resource_group'] = id_dict.get('resource_group')

        return d


def main():
    AzureRMApplicationFirewallPolicy()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_vrf_region
short_description: Manage site-local VRF regions in schema template
description:
- Manage site-local VRF regions in schema template on Cisco ACI Multi-Site.
author:
- Anvitha Jain (@anvitha-jain)
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  vrf:
    description:
    - The name of the VRF.
    type: str
    required: true
  region:
    description:
    - The name of the region to manage.
    type: str
    aliases: [ name ]
  vpn_gateway_router:
    description:
    - Whether VPN Gateway Router is enabled or not.
    type: bool
  container_overlay:
    description:
    - The name of the context profile type.
    - This is supported on versions of MSO that are 3.3 or greater.
    type: bool
  underlay_context_profile:
    description:
    - The name of the context profile type.
    - This parameter can only be added when container_overlay is True.
    - This is supported on versions of MSO that are 3.3 or greater.
    type: dict
    suboptions:
      vrf:
        description:
        - The name of the VRF to associate with underlay context profile.
        type: str
        required: true
      region:
        description:
        - The name of the region associated with underlay context profile VRF.
        type: str
        required: true
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- Due to restrictions of the MSO REST API, this module cannot create empty region (i.e. regions without cidrs)
  Use the M(cisco.mso.mso_schema_site_vrf_region_cidr) to automatically create regions with cidrs.
seealso:
- module: cisco.mso.mso_schema_site_vrf
- module: cisco.mso.mso_schema_template_vrf
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Remove VPN Gateway Router at site VRF Region
  cisco.mso.mso_schema_site_vrf_region:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    vpn_gateway_router: false
    state: present

- name: Remove a site VRF region
  cisco.mso.mso_schema_site_vrf_region:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    state: absent

- name: Query a specific site VRF region
  cisco.mso.mso_schema_site_vrf_region:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    state: query
  register: query_result

- name: Query all site VRF regions
  cisco.mso.mso_schema_site_vrf_region:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        vrf=dict(type="str", required=True),
        region=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        vpn_gateway_router=dict(type="bool"),
        container_overlay=dict(type="bool"),
        underlay_context_profile=dict(
            type="dict",
            options=dict(
                vrf=dict(type="str", required=True),
                region=dict(type="str", required=True),
            ),
        ),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["region"]],
            ["state", "present", ["region"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    vrf = module.params.get("vrf")
    region = module.params.get("region")
    vpn_gateway_router = module.params.get("vpn_gateway_router")
    container_overlay = module.params.get("container_overlay")
    underlay_context_profile = module.params.get("underlay_context_profile")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(msg="Provided site-template association '{0}-{1}' does not exist.".format(site, template))

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    # Get VRF
    vrf_ref = mso.vrf_ref(schema_id=schema_id, template=template, vrf=vrf)
    vrfs = [v.get("vrfRef") for v in schema_obj.get("sites")[site_idx]["vrfs"]]
    vrfs_name = [mso.dict_from_ref(v).get("vrfName") for v in vrfs]
    if vrf_ref not in vrfs:
        mso.fail_json(msg="Provided vrf '{0}' does not exist. Existing vrfs: {1}".format(vrf, ", ".join(vrfs_name)))
    vrf_idx = vrfs.index(vrf_ref)

    # Get Region
    regions = [r.get("name") for r in schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"]]
    if region is not None and region in regions:
        region_idx = regions.index(region)
        region_path = "/sites/{0}/vrfs/{1}/regions/{2}".format(site_template, vrf, region)
        mso.existing = schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]

    if state == "query":
        if region is None:
            mso.existing = schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"]
        elif not mso.existing:
            mso.fail_json(msg="Region '{region}' not found".format(region=region))
        mso.exit_json()

    regions_path = "/sites/{0}/vrfs/{1}/regions".format(site_template, vrf)
    ops = []

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=region_path))

    elif state == "present":
        payload = dict(
            name=region,
            isVpnGatewayRouter=vpn_gateway_router,
        )

        if container_overlay:
            payload["contextProfileType"] = "container-overlay"
            if mso.existing:
                underlay_dict = dict(
                    vrfRef=dict(schemaId=schema_id, templateName=template, vrfName=underlay_context_profile["vrf"]),
                    regionName=underlay_context_profile["region"],
                )
                payload["underlayCtxProfile"] = underlay_dict

        mso.sanitize(payload, collate=True)
        if mso.existing:
            ops.append(dict(op="replace", path=region_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=regions_path + "/-", value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: linode_v4
short_description: Manage instances on the Linode cloud
description: Manage instances on the Linode cloud.
requirements:
  - linode_api4 >= 2.0.0
author:
  - Luke Murphy (@decentral1se)
notes:
  - No Linode resizing is currently implemented. This module aims to replace the current Linode module which uses deprecated
    API bindings on the Linode side.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
options:
  region:
    description:
      - The region of the instance. This is a required parameter only when creating Linode instances. See U(https://www.linode.com/docs/api/regions/).
    type: str
  image:
    description:
      - The image of the instance. This is a required parameter only when creating Linode instances.
      - See U(https://www.linode.com/docs/api/images/).
    type: str
  type:
    description:
      - The type of the instance. This is a required parameter only when creating Linode instances.
      - See U(https://www.linode.com/docs/api/linode-types/).
    type: str
  label:
    description:
      - The instance label. This label is used as the main determiner for idempotency for the module and is therefore mandatory.
    type: str
    required: true
  group:
    description:
      - The group that the instance should be marked under. Please note, that group labelling is deprecated but still supported.
        The encouraged method for marking instances is to use tags.
    type: str
  private_ip:
    description:
      - If V(true), the created Linode instance has private networking enabled and assigned a private IPv4 address.
    type: bool
    default: false
    version_added: 3.0.0
  tags:
    description:
      - The tags that the instance should be marked under.
      - See U(https://www.linode.com/docs/api/tags/).
    type: list
    elements: str
  root_pass:
    description:
      - The password for the root user. If not specified, it generates a new one. This generated password is available in
        the task success JSON.
    type: str
  authorized_keys:
    description:
      - A list of SSH public key parts to deploy for the root user.
    type: list
    elements: str
  state:
    description:
      - The desired instance state.
    type: str
    choices:
      - present
      - absent
    required: true
  access_token:
    description:
      - The Linode API v4 access token. It may also be specified by exposing the E(LINODE_ACCESS_TOKEN) environment variable.
      - See U(https://www.linode.com/docs/api#access-and-authentication).
    required: true
    type: str
  stackscript_id:
    description:
      - The numeric ID of the StackScript to use when creating the instance.
      - See U(https://www.linode.com/docs/api/stackscripts/).
    type: int
    version_added: 1.3.0
  stackscript_data:
    description:
      - An object containing arguments to any User Defined Fields present in the StackScript used when creating the instance.
        Only valid when a O(stackscript_id) is provided.
      - See U(https://www.linode.com/docs/api/stackscripts/).
    type: dict
    version_added: 1.3.0
"""

EXAMPLES = r"""
- name: Create a new Linode.
  community.general.linode_v4:
    label: new-linode
    type: g6-nanode-1
    region: eu-west
    image: linode/debian9
    root_pass: passw0rd
    authorized_keys:
      - "ssh-rsa ..."
    stackscript_id: 1337
    stackscript_data:
      variable: value
    state: present

- name: Delete that new Linode.
  community.general.linode_v4:
    label: new-linode
    state: absent
"""

RETURN = r"""
instance:
  description: The instance description in JSON serialized form.
  returned: Always.
  type: dict
  sample:
    {
      "root_pass": "foobar",  # if auto-generated
      "alerts": {
        "cpu": 90,
        "io": 10000,
        "network_in": 10,
        "network_out": 10,
        "transfer_quota": 80
      },
      "backups": {
        "enabled": false,
        "schedule": {
          "day": null,
          "window": null
        }
      },
      "created": "2018-09-26T08:12:33",
      "group": "Foobar Group",
      "hypervisor": "kvm",
      "id": 10480444,
      "image": "linode/centos7",
      "ipv4": [
        "130.132.285.233"
      ],
      "ipv6": "2a82:7e00::h03c:46ff:fe04:5cd2/64",
      "label": "lin-foo",
      "region": "eu-west",
      "specs": {
        "disk": 25600,
        "memory": 1024,
        "transfer": 1000,
        "vcpus": 1
      },
      "status": "running",
      "tags": [],
      "type": "g6-nanode-1",
      "updated": "2018-09-26T10:10:14",
      "watchdog_enabled": true
    }
"""

import traceback

from ansible.module_utils.basic import AnsibleModule, env_fallback, missing_required_lib

from ansible_collections.community.general.plugins.module_utils._linode import get_user_agent

LINODE_IMP_ERR = None
try:
    from linode_api4 import Instance, LinodeClient

    HAS_LINODE_DEPENDENCY = True
except ImportError:
    LINODE_IMP_ERR = traceback.format_exc()
    HAS_LINODE_DEPENDENCY = False


def create_linode(module, client, **kwargs):
    """Creates a Linode instance and handles return format."""
    if kwargs["root_pass"] is None:
        kwargs.pop("root_pass")

    try:
        response = client.linode.instance_create(**kwargs)
    except Exception as exception:
        module.fail_json(msg=f"Unable to query the Linode API. Saw: {exception}")

    try:
        if isinstance(response, tuple):
            instance, root_pass = response
            instance_json = instance._raw_json
            instance_json.update({"root_pass": root_pass})
            return instance_json
        else:
            return response._raw_json
    except TypeError:
        module.fail_json(
            msg="Unable to parse Linode instance creation response. Please raise a bug against this"
            " module on https://github.com/ansible-collections/community.general/issues"
        )


def maybe_instance_from_label(module, client):
    """Try to retrieve an instance based on a label."""
    try:
        label = module.params["label"]
        result = client.linode.instances(Instance.label == label)
        return result[0]
    except IndexError:
        return None
    except Exception as exception:
        module.fail_json(msg=f"Unable to query the Linode API. Saw: {exception}")


def initialise_module():
    """Initialise the module parameter specification."""
    return AnsibleModule(
        argument_spec=dict(
            label=dict(type="str", required=True),
            state=dict(type="str", required=True, choices=["present", "absent"]),
            access_token=dict(
                type="str",
                required=True,
                no_log=True,
                fallback=(env_fallback, ["LINODE_ACCESS_TOKEN"]),
            ),
            authorized_keys=dict(type="list", elements="str", no_log=False),
            group=dict(type="str"),
            image=dict(type="str"),
            private_ip=dict(type="bool", default=False),
            region=dict(type="str"),
            root_pass=dict(type="str", no_log=True),
            tags=dict(type="list", elements="str"),
            type=dict(type="str"),
            stackscript_id=dict(type="int"),
            stackscript_data=dict(type="dict"),
        ),
        supports_check_mode=False,
        required_one_of=(["state", "label"],),
        required_together=(["region", "image", "type"],),
    )


def build_client(module):
    """Build a LinodeClient."""
    return LinodeClient(module.params["access_token"], user_agent=get_user_agent("linode_v4_module"))


def main():
    """Module entrypoint."""
    module = initialise_module()

    if not HAS_LINODE_DEPENDENCY:
        module.fail_json(msg=missing_required_lib("linode-api4"), exception=LINODE_IMP_ERR)

    client = build_client(module)
    instance = maybe_instance_from_label(module, client)

    if module.params["state"] == "present" and instance is not None:
        module.exit_json(changed=False, instance=instance._raw_json)

    elif module.params["state"] == "present" and instance is None:
        instance_json = create_linode(
            module,
            client,
            authorized_keys=module.params["authorized_keys"],
            group=module.params["group"],
            image=module.params["image"],
            label=module.params["label"],
            private_ip=module.params["private_ip"],
            region=module.params["region"],
            root_pass=module.params["root_pass"],
            tags=module.params["tags"],
            ltype=module.params["type"],
            stackscript=module.params["stackscript_id"],
            stackscript_data=module.params["stackscript_data"],
        )
        module.exit_json(changed=True, instance=instance_json)

    elif module.params["state"] == "absent" and instance is not None:
        instance.delete()
        module.exit_json(changed=True, instance=instance._raw_json)

    elif module.params["state"] == "absent" and instance is None:
        module.exit_json(changed=False, instance={})


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2021 Aparna Patil(@aparna-patil)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_hostgroup

version_added: "1.10.0"

short_description: Create, delete and update a dedicated host group

description:
    - Creates, deletes, and updates a dedicated host group.

options:
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    name:
        description:
            - The name of the dedicated host group.
        required: true
        type: str
    location:
        description:
            - Valid Azure location for host group. Defaults to location of resource group.
        type: str
    platform_fault_domain_count:
        description:
            - Number of fault domains that the host group can span.
        type: int
    zones:
        description:
            - Availability Zone to use for this host group. Only single zone is supported. The zone can be assigned only
             during creation. If not provided, the group supports all zones in the region.
        type: list
        elements: str
    state:
        description:
            - Assert the state of the host group. Use C(present) to create or update and C(absent) to delete.
        default: present
        type: str
        choices:
            - absent
            - present

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Aparna Patil (@aparna-patil)
'''

EXAMPLES = '''
- name: Create a host group
  azure_rm_hostgroup:
    resource_group: myAzureResourceGroup
    name: myhostgroup
    location: eastus
    zones:
      - "1"
    platform_fault_domain_count: 1
    state: present

- name: Update a host group
  azure_rm_hostgroup:
    resource_group: myAzureResourceGroup
    name: myhostgroup
    location: eastus
    zones:
      - "1"
    platform_fault_domain_count: 1
    state: present
    tags:
      key1: "value1"

- name: Delete a host group
  azure_rm_hostgroup:
    resource_group: myAzureResourceGroup
    name: myhostgroup
    state: absent
'''

RETURN = '''
state:
    description:
        - Current state of the host group.
    returned: always
    type: complex
    contains:
        id:
            description:
                - The host group ID.
            returned: always
            type: str
            sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myAzureResourceGroup/providers/
                     Microsoft.Compute/hostGroups/myhostgroup"
        name:
            description:
                - The host group name.
            returned: always
            type: str
            sample: 'myhostgroup'
        location:
            description:
                - The Azure Region where the resource lives.
            returned: always
            type: str
            sample: eastus
        platform_fault_domain_count:
            description:
                - Number of fault domains.
            returned: always
            type: int
            sample: 1
        zones:
            description:
                - Availability zones configured for this host group.
            returned: always
            type: list
            sample: ["1"]
        tags:
            description:
                - Resource tags.
            returned: always
            type: list
            sample: [{"key1": "value1"}]
        type:
            description:
                - The type of resource.
            returned: always
            type: str
            sample: Microsoft.Compute/hostGroups
'''

from ansible.module_utils.basic import _load_params
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase, \
    normalize_location_name

try:
    from azure.core.exceptions import ResourceNotFoundError
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMHostGroup(AzureRMModuleBase):

    def __init__(self):

        _load_params()
        # define user inputs from playbook
        self.module_arg_spec = dict(
            resource_group=dict(type='str', required=True),
            name=dict(type='str', required=True),
            location=dict(type='str'),
            platform_fault_domain_count=dict(type='int'),
            zones=dict(type='list', elements='str'),
            state=dict(choices=['present', 'absent'], default='present', type='str')
        )

        required_if = [
            ('state', 'present', ['platform_fault_domain_count'])
        ]

        self.results = dict(
            changed=False,
            state=dict()
        )

        self.resource_group = None
        self.name = None
        self.state = None
        self.location = None
        self.tags = None
        self.platform_fault_domain_count = None
        self.zones = None

        super(AzureRMHostGroup, self).__init__(self.module_arg_spec,
                                               required_if=required_if,
                                               supports_check_mode=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False
        results = dict()
        host_group = None

        # retrieve resource group to make sure it exists
        resource_group = self.get_resource_group(self.resource_group)
        if not self.location:
            # Set default location
            self.location = resource_group.location

        self.location = normalize_location_name(self.location)

        try:
            self.log('Fetching host group {0}'.format(self.name))
            host_group = self.compute_client.dedicated_host_groups.get(self.resource_group, self.name)
            # serialize object into a dictionary
            results = self.hostgroup_to_dict(host_group)
            if self.state == 'present':
                changed = False
                update_tags, results['tags'] = self.update_tags(results['tags'])
                if update_tags:
                    changed = True
                self.tags = results['tags']
                if self.platform_fault_domain_count != results['platform_fault_domain_count']:
                    self.fail("Error updating host group : {0}. Changing platform_fault_domain_count is not allowed."
                              .format(self.name))
                if self.zones:
                    if ('zones' in results and self.zones[0] != results['zones'][0]) or 'zones' not in results:
                        self.fail("Error updating host group : {0}. Changing property zones is not allowed."
                                  .format(self.name))
            elif self.state == 'absent':
                changed = True

        except ResourceNotFoundError:
            if self.state == 'present':
                changed = True
            else:
                changed = False

        self.results['changed'] = changed
        self.results['state'] = results

        if self.check_mode:
            return self.results

        if changed:
            if self.state == 'present':
                # create or update a dedicated host group
                host_group_new = \
                    self.compute_models.DedicatedHostGroup(location=self.location,
                                                           platform_fault_domain_count=self.platform_fault_domain_count,
                                                           zones=self.zones)
                if self.tags:
                    host_group_new.tags = self.tags
                self.results['state'] = self.create_or_update_hostgroup(host_group_new)

            elif self.state == 'absent':
                # delete a host group
                self.delete_hostgroup()
                self.results['state'] = 'Deleted'

        return self.results

    def create_or_update_hostgroup(self, host_group):
        try:
            # create the host group
            response = self.compute_client.dedicated_host_groups.create_or_update(
                resource_group_name=self.resource_group,
                host_group_name=self.name,
                parameters=host_group)
        except Exception as exc:
            self.fail("Error creating or updating host group {0} - {1}".format(self.name, str(exc)))
        return self.hostgroup_to_dict(response)

    def delete_hostgroup(self):
        try:
            # delete the host group
            response = self.compute_client.dedicated_host_groups.delete(resource_group_name=self.resource_group,
                                                                        host_group_name=self.name)
        except Exception as exc:
            self.fail("Error deleting host group {0} - {1}".format(self.name, str(exc)))
        return response

    def hostgroup_to_dict(self, hostgroup):
        result = hostgroup.as_dict()
        result['tags'] = hostgroup.tags
        return result


def main():
    AzureRMHostGroup()


if __name__ == '__main__':
    main()



# Copyright (c) 2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import traceback
import typing as t

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_native, to_text

try:
    import dns
    import dns.exception
    import dns.inet
    import dns.message
    import dns.name
    import dns.query
    import dns.rcode
    import dns.rdatatype
    import dns.resolver
    import dns.version
except ImportError:
    DNSPYTHON_IMPORTERROR = traceback.format_exc()
else:
    DNSPYTHON_IMPORTERROR = None  # type: ignore  # TODO

if t.TYPE_CHECKING:
    from collections.abc import Collection, Mapping, Sequence  # pragma: no cover

    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    _P = t.ParamSpec("_P")  # pragma: no cover
    _T = t.TypeVar("_T")  # pragma: no cover

    class ResolverParams(t.TypedDict):  # pragma: no cover
        search: t.NotRequired[bool]  # pragma: no cover


_EDNS_SIZE = 1232  # equals dns.message.DEFAULT_EDNS_PAYLOAD; larger values cause problems with Route53 nameservers for me


class ResolverError(Exception):
    pass


class InvalidInput(ResolverError):
    pass


class _Resolve:
    def __init__(
        self, timeout: float = 10, timeout_retries: int = 3, servfail_retries: int = 0
    ) -> None:
        self.timeout = timeout
        self.timeout_retries = timeout_retries
        self.servfail_retries = servfail_retries
        self.default_resolver = dns.resolver.get_default_resolver()

    def _handle_reponse_errors(
        self,
        target: dns.name.Name,
        response: dns.message.Message,
        nameserver: (
            Sequence[str | dns.nameserver.Nameserver]
            | str
            | dns.nameserver.Nameserver
            | None
        ) = None,
        query: str | None = None,
        accept_errors: Collection[dns.rcode.Rcode] | None = None,
    ):
        rcode = response.rcode()
        if rcode == dns.rcode.NOERROR:
            return True
        if accept_errors and rcode in accept_errors:
            return True
        if rcode == dns.rcode.NXDOMAIN:
            raise dns.resolver.NXDOMAIN(qnames=[target], responses={target: response})
        msg = f"Error {dns.rcode.to_text(rcode)}"
        if nameserver:
            msg = f"{msg} while querying {nameserver}"
        if query:
            msg = f"{msg} with query {query}"
        raise ResolverError(msg)

    def _handle_timeout(
        self, function: t.Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs
    ) -> _T:
        retry = 0
        while True:
            try:
                return function(*args, **kwargs)
            except dns.exception.Timeout as exc:
                if retry >= self.timeout_retries:
                    raise exc
                retry += 1

    def _resolve(
        self,
        resolver: dns.resolver.Resolver,
        dnsname: dns.name.Name,
        *,
        handle_response_errors: bool = False,
        rdtype: dns.rdatatype.RdataType,
        **kwargs: t.Unpack[ResolverParams],
    ) -> dns.rrset.RRset | None:
        retry = 0
        while True:
            response = self._handle_timeout(
                resolver.resolve,
                dnsname,
                lifetime=self.timeout,
                rdtype=rdtype,
                **kwargs,
            )
            if (
                response.response.rcode() == dns.rcode.SERVFAIL
                and retry < self.servfail_retries
            ):
                retry += 1
                continue
            if handle_response_errors:
                self._handle_reponse_errors(
                    dnsname, response.response, nameserver=resolver.nameservers
                )
            return response.rrset


class SimpleResolver(_Resolve):
    def __init__(
        self,
        timeout: float = 10,
        timeout_retries: int = 3,
        servfail_retries: int = 0,
    ) -> None:
        super().__init__(
            timeout=timeout,
            timeout_retries=timeout_retries,
            servfail_retries=servfail_retries,
        )

    def resolve(
        self,
        target: str,
        *,
        nxdomain_is_empty: bool = True,
        server_addresses=None,
        target_can_be_relative: bool = False,
        rdtype: dns.rdatatype.RdataType,
        **kwargs: t.Unpack[ResolverParams]
    ):
        dnsname = (
            dns.name.from_unicode(to_text(target), origin=None)
            if target_can_be_relative
            else dns.name.from_unicode(to_text(target))
        )

        resolver = self.default_resolver
        if server_addresses:
            resolver = dns.resolver.Resolver(configure=False)
            resolver.timeout = self.timeout
            resolver.nameservers = server_addresses

        resolver.use_edns(0, ednsflags=dns.flags.DO, payload=_EDNS_SIZE)

        try:
            return self._resolve(
                resolver, dnsname, handle_response_errors=True, rdtype=rdtype, **kwargs
            )
        except dns.resolver.NXDOMAIN:
            if nxdomain_is_empty:
                return None
            raise
        except dns.resolver.NoAnswer:
            return None

    def resolve_addresses(
        self, target: str | bytes, **kwargs: t.Unpack[ResolverParams]
    ) -> list[str]:
        dnsname = dns.name.from_unicode(to_text(target))
        resolver = self.default_resolver
        result = []
        try:
            for data in (
                self._resolve(
                    resolver,
                    dnsname,
                    handle_response_errors=True,
                    rdtype=dns.rdatatype.A,
                    **kwargs,
                )
                or ()
            ):
                result.append(str(data))
        except dns.resolver.NoAnswer:
            pass
        try:
            for data in (
                self._resolve(
                    resolver,
                    dnsname,
                    handle_response_errors=True,
                    rdtype=dns.rdatatype.AAAA,
                    **kwargs,
                )
                or ()
            ):
                result.append(str(data))
        except dns.resolver.NoAnswer:
            pass
        return result


class ResolveDirectlyFromNameServers(_Resolve):
    def __init__(
        self,
        timeout: float = 10,
        timeout_retries: int = 3,
        servfail_retries: int = 0,
        always_ask_default_resolver: bool = True,
        server_addresses: Sequence[str] | None = None,
    ) -> None:
        super().__init__(
            timeout=timeout,
            timeout_retries=timeout_retries,
            servfail_retries=servfail_retries,
        )
        self.cache: dict[tuple[str, t.Literal["ns", "addr"]], list[str]] = {}
        self.cname_cache: dict[str, dns.name.Name | None] = {}
        self.resolver_cache: dict[str, dns.resolver.Resolver] = {}
        self.default_nameservers: list[str | dns.nameserver.Nameserver] = list(
            self.default_resolver.nameservers
            if server_addresses is None
            else server_addresses
        )
        self.always_ask_default_resolver = always_ask_default_resolver

    def _lookup_ns_names(
        self,
        target: dns.name.Name,
        nameservers: Sequence[str] | None = None,
        nameserver_ips: Sequence[str | dns.nameserver.Nameserver] | None = None,
    ) -> tuple[list[str] | None, dns.name.Name | None]:
        if self.always_ask_default_resolver:
            nameservers = None
            nameserver_ips = self.default_nameservers
        if nameservers is None and nameserver_ips is None:
            nameserver_ips = self.default_nameservers
        if not nameserver_ips and nameservers:
            for ns in nameservers:
                nameserver_ips = self._lookup_address(ns)
                if nameserver_ips:
                    break
        if not nameserver_ips:
            if nameservers:
                raise ResolverError(
                    "Have neither nameservers nor nameserver IPs: the given nameservers do not resolve to IPs"
                )
            raise ResolverError("Have neither nameservers nor nameserver IPs")

        nameserver = nameserver_ips[0]

        # Sanity check: do we have a valid nameserver IP?
        if isinstance(nameserver, str):
            try:
                dns.inet.af_for_address(nameserver)
            except ValueError as exc:
                raise InvalidInput(
                    f"Invalid nameserver IP address {nameserver}"
                ) from exc

        query = dns.message.make_query(target, dns.rdatatype.NS)
        retry = 0
        while True:
            if isinstance(nameserver, str):
                response = self._handle_timeout(
                    dns.query.udp, query, nameserver, timeout=self.timeout
                )
            else:
                response = self._handle_timeout(
                    nameserver.query,
                    query,
                    timeout=self.timeout,
                    # The following are taken from the default arguments of
                    # dns.resolver.Resolver.resolve():
                    source=None,
                    source_port=0,
                    max_size=False,
                )
            if response.rcode() == dns.rcode.SERVFAIL and retry < self.servfail_retries:
                retry += 1
                continue
            break
        self._handle_reponse_errors(
            target,
            response,
            nameserver=nameserver,
            query=f'get NS for "{target}"',
            accept_errors=[dns.rcode.NXDOMAIN],
        )

        cname = None
        for rrset in response.answer:
            if rrset.rdtype == dns.rdatatype.CNAME:
                cname = dns.name.from_text(to_text(rrset[0]))

        new_nameservers: list[str] = []
        rrsets = list(response.authority)
        rrsets.extend(response.answer)
        for rrset in rrsets:
            if rrset.rdtype == dns.rdatatype.SOA:
                # We keep the current nameservers
                return None, cname
            if rrset.rdtype == dns.rdatatype.NS:
                new_nameservers.extend(str(ns_record.target) for ns_record in rrset)
        return sorted(set(new_nameservers)) if new_nameservers else None, cname

    def _lookup_address_impl(
        self, target: dns.name.Name, rdtype: dns.rdatatype.RdataType
    ) -> list[str]:
        try:
            answer = self._resolve(
                self.default_resolver,
                target,
                handle_response_errors=True,
                rdtype=rdtype,
            )
            return [str(res) for res in answer or ()]
        except dns.resolver.NoAnswer:
            return []

    def _lookup_address(self, target) -> list[str]:
        result = self.cache.get((target, "addr"))
        if result is None:
            result = self._lookup_address_impl(target, dns.rdatatype.A)
            result.extend(self._lookup_address_impl(target, dns.rdatatype.AAAA))
            self.cache[(target, "addr")] = result
        return result

    def _do_lookup_ns(self, target: dns.name.Name) -> list[str] | None:
        nameserver_ips: Sequence[str | dns.nameserver.Nameserver] | None = (
            self.default_nameservers
        )
        nameservers: list[str] | None = None
        for i in range(2, len(target.labels) + 1):
            target_part = target.split(i)[1]
            _nameservers = self.cache.get((str(target_part), "ns"))
            if _nameservers is None:
                nameserver_names, cname = self._lookup_ns_names(
                    target_part, nameservers=nameservers, nameserver_ips=nameserver_ips
                )
                if nameserver_names is not None:
                    nameservers = nameserver_names

                if nameservers is not None:
                    self.cache[(str(target_part), "ns")] = nameservers
                self.cname_cache[str(target_part)] = cname
            else:
                nameservers = _nameservers
            nameserver_ips = None

        return nameservers

    def _lookup_ns(self, target: dns.name.Name) -> list[str] | None:
        result = self.cache.get((str(target), "ns"))
        if result is None:
            result = self._do_lookup_ns(target)
            if result is not None:
                self.cache[(str(target), "ns")] = result
        return result

    def _get_resolver(
        self, dnsname: dns.name.Name, nameservers
    ) -> dns.resolver.Resolver:
        cache_index = "|".join([str(dnsname)] + sorted(nameservers))
        resolver = self.resolver_cache.get(cache_index)
        if resolver is None:
            resolver = dns.resolver.Resolver(configure=False)
            resolver.use_edns(0, ednsflags=dns.flags.DO, payload=_EDNS_SIZE)
            resolver.timeout = self.timeout
            nameserver_ips = set()
            for nameserver in nameservers:
                nameserver_ips.update(self._lookup_address(nameserver))
            resolver.nameservers = sorted(nameserver_ips)
            self.resolver_cache[cache_index] = resolver
        return resolver

    def resolve_nameservers(
        self, target: str | bytes, resolve_addresses: bool = False
    ) -> list[str]:
        nameservers = self._lookup_ns(dns.name.from_unicode(to_text(target)))
        if resolve_addresses:
            nameserver_ips = set()
            for nameserver in nameservers or []:
                nameserver_ips.update(self._lookup_address(nameserver))
            nameservers = list(nameserver_ips)
        return sorted(nameservers or [])

    def resolve(
        self,
        target: str | bytes,
        *,
        nxdomain_is_empty: bool = True,
        rdtype: dns.rdatatype.RdataType,
        **kwargs: t.Unpack[ResolverParams]
    ) -> dict[str, dns.rrset.RRset | None]:
        dnsname = dns.name.from_unicode(to_text(target))
        loop_catcher = set()
        while True:
            try:
                nameservers = self._lookup_ns(dnsname)
            except dns.resolver.NXDOMAIN:
                if nxdomain_is_empty:
                    return {}
                raise
            cname = self.cname_cache.get(str(dnsname))
            if cname is None:
                break
            dnsname = cname
            if dnsname in loop_catcher:
                raise ResolverError(f"Found CNAME loop starting at {to_native(target)}")
            loop_catcher.add(dnsname)

        results: dict[str, dns.rrset.RRset | None] = {}
        for nameserver in nameservers or []:
            results[nameserver] = None
            resolver = self._get_resolver(dnsname, [nameserver])
            try:
                results[nameserver] = self._resolve(
                    resolver,
                    dnsname,
                    handle_response_errors=True,
                    rdtype=rdtype,
                    **kwargs,
                )
            except dns.resolver.NoAnswer:
                pass
            except dns.resolver.NXDOMAIN:
                if nxdomain_is_empty:
                    # Note that rdclass is not always correct, but it's good enough for us...
                    results[nameserver] = dns.rrset.RRset(
                        name=dnsname, rdclass=dns.rdataclass.IN, rdtype=rdtype
                    )
                else:
                    raise
        return results


def guarded_run(
    runner: t.Callable[[], _T],
    module: AnsibleModule,
    *,
    server: str | None = None,
    generate_additional_results: t.Callable[[], Mapping[str, t.Any]] | None = None,
) -> _T:
    suffix = f" for {server}" if server is not None else ""
    kwargs: Mapping[str, t.Any] = {}
    try:
        return runner()
    except InvalidInput as e:
        if generate_additional_results is not None:
            kwargs = generate_additional_results()
        module.fail_json(
            msg=f"Invalid input{suffix}: {to_native(e)}",
            exception=traceback.format_exc(),
            **kwargs,
        )
    except ResolverError as e:
        if generate_additional_results is not None:
            kwargs = generate_additional_results()
        module.fail_json(
            msg=f"Unexpected resolving error{suffix}: {to_native(e)}",
            exception=traceback.format_exc(),
            **kwargs,
        )
    except dns.exception.DNSException as e:
        if generate_additional_results is not None:
            kwargs = generate_additional_results()
        module.fail_json(
            msg=f"Unexpected DNS error{suffix}: {to_native(e)}",
            exception=traceback.format_exc(),
            **kwargs,
        )


def assert_requirements_present(module: AnsibleModule) -> None:
    if DNSPYTHON_IMPORTERROR is not None:
        module.fail_json(
            msg=missing_required_lib("dnspython >= 2.0.0"),
            exception=DNSPYTHON_IMPORTERROR,
        )
    if dns.version.MAJOR < 2:
        module.fail_json(
            msg=missing_required_lib("dnspython >= 2.0.0")
            + f" Found version {dns.version.version}.",
            exception=DNSPYTHON_IMPORTERROR,
        )



# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import base64
import binascii
import datetime
import ipaddress
import os
import re
import tempfile
import traceback
import typing as t

from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._acme.backends import (
    CertificateInformation,
    CryptoBackend,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    BackendException,
    KeyParsingError,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.utils import (
    nopad_b64,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.math import (
    convert_bytes_to_int,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    ensure_utc_timezone,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.module_utils._acme.certificates import (  # pragma: no cover
        Criterium,
    )


_OPENSSL_ENVIRONMENT_UPDATE = {
    "LANG": "C",
    "LC_ALL": "C",
    "LC_MESSAGES": "C",
    "LC_CTYPE": "C",
}


def _extract_date(
    out_text: str, *, name: str, cert_filename_suffix: str = ""
) -> datetime.datetime:
    matcher = re.search(rf"\s+{name}\s*:\s+(.*)", out_text)
    if matcher is None:
        raise BackendException(f"No '{name}' date found{cert_filename_suffix}")
    date_str = matcher.group(1)
    try:
        # For some reason Python's strptime() does not return any timezone information,
        # even though the information is there and a supported timezone for all supported
        # Python implementations (GMT). So we have to modify the datetime object by
        # replacing it by UTC.
        return ensure_utc_timezone(
            datetime.datetime.strptime(date_str, "%b %d %H:%M:%S %Y %Z")
        )
    except ValueError as exc:
        raise BackendException(
            f"Failed to parse '{name}' date{cert_filename_suffix}: {exc}"
        ) from exc


def _decode_octets(octets_text: str) -> bytes:
    return binascii.unhexlify(re.sub(r"(\s|:)", "", octets_text).encode("utf-8"))


@t.overload
def _extract_octets(
    out_text: str,
    *,
    name: str,
    required: t.Literal[False],
    potential_prefixes: t.Iterable[str] | None = None,
) -> bytes | None: ...


@t.overload
def _extract_octets(
    out_text: str,
    *,
    name: str,
    required: t.Literal[True],
    potential_prefixes: t.Iterable[str] | None = None,
) -> bytes: ...


def _extract_octets(
    out_text: str,
    *,
    name: str,
    required: bool = True,
    potential_prefixes: t.Iterable[str] | None = None,
) -> bytes | None:
    part = (
        f"(?:{'|'.join(re.escape(pp) for pp in potential_prefixes)})"
        if potential_prefixes
        else ""
    )
    regexp = rf"\s+{name}:\s*\n\s+{part}([A-Fa-f0-9]{{2}}(?::[A-Fa-f0-9]{{2}})*)\s*\n"
    match = re.search(regexp, out_text, re.MULTILINE | re.DOTALL)
    if match is not None:
        return _decode_octets(match.group(1))
    if not required:
        return None
    raise BackendException(f"No '{name}' octet string found")


def _extract_rsa_key(out_text: str, *, key_file: str | None = None) -> dict[str, t.Any]:
    matcher = re.search(
        r"modulus:\n\s+(?:00:)?([a-f0-9\:\s]+?)\npublicExponent",
        out_text,
        re.MULTILINE | re.DOTALL,
    )
    if matcher is None:
        raise KeyParsingError("cannot parse RSA key: modulus not found")
    pub_hex = matcher.group(1)

    matcher = re.search(
        r"\npublicExponent: ([0-9]+)", out_text, re.MULTILINE | re.DOTALL
    )
    if matcher is None:
        raise KeyParsingError("cannot parse RSA key: public exponent not found")
    pub_exp = matcher.group(1)
    pub_exp = f"{int(pub_exp):x}"
    if len(pub_exp) % 2:
        pub_exp = f"0{pub_exp}"

    return {
        "key_file": key_file,
        "type": "rsa",
        "alg": "RS256",
        "jwk": {
            "kty": "RSA",
            "e": nopad_b64(binascii.unhexlify(pub_exp.encode("utf-8"))),
            "n": nopad_b64(_decode_octets(pub_hex)),
        },
        "hash": "sha256",
    }


def _extract_ec_key(out_text: str, *, key_file: str | None = None) -> dict[str, t.Any]:
    pub_data = re.search(
        r"pub:\s*\n\s+04:([a-f0-9\:\s]+?)\nASN1 OID: (\S+)(?:\nNIST CURVE: (\S+))?",
        out_text,
        re.MULTILINE | re.DOTALL,
    )
    if pub_data is None:
        raise KeyParsingError("cannot parse elliptic curve key")
    pub_hex = _decode_octets(pub_data.group(1))
    asn1_oid_curve = pub_data.group(2).lower()
    nist_curve = pub_data.group(3).lower() if pub_data.group(3) else None
    if asn1_oid_curve == "prime256v1" or nist_curve == "p-256":
        bits = 256
        alg = "ES256"
        hashalg = "sha256"
        point_size = 32
        curve = "P-256"
    elif asn1_oid_curve == "secp384r1" or nist_curve == "p-384":
        bits = 384
        alg = "ES384"
        hashalg = "sha384"
        point_size = 48
        curve = "P-384"
    elif asn1_oid_curve == "secp521r1" or nist_curve == "p-521":
        # Not yet supported on Let's Encrypt side, see
        # https://github.com/letsencrypt/boulder/issues/2217
        bits = 521
        alg = "ES512"
        hashalg = "sha512"
        point_size = 66
        curve = "P-521"
    else:
        raise KeyParsingError(
            f"unknown elliptic curve: {asn1_oid_curve} / {nist_curve}"
        )
    num_bytes = (bits + 7) // 8
    if len(pub_hex) != 2 * num_bytes:
        raise KeyParsingError(
            f"bad elliptic curve point ({asn1_oid_curve} / {nist_curve})"
        )
    return {
        "key_file": key_file,
        "type": "ec",
        "alg": alg,
        "jwk": {
            "kty": "EC",
            "crv": curve,
            "x": nopad_b64(pub_hex[:num_bytes]),
            "y": nopad_b64(pub_hex[num_bytes:]),
        },
        "hash": hashalg,
        "point_size": point_size,
    }


class OpenSSLCLIBackend(CryptoBackend):
    def __init__(
        self, *, module: AnsibleModule, openssl_binary: str | None = None
    ) -> None:
        super().__init__(module=module, with_timezone=True)
        if openssl_binary is None:
            openssl_binary = module.get_bin_path("openssl", True)
        self.openssl_binary = openssl_binary

    def parse_key(
        self,
        *,
        key_file: str | os.PathLike | None = None,
        key_content: str | None = None,
        passphrase: str | None = None,
    ) -> dict[str, t.Any]:
        """
        Parses an RSA or Elliptic Curve key file in PEM format and returns key_data.
        Raises KeyParsingError in case of errors.
        """
        if passphrase is not None:
            raise KeyParsingError("openssl backend does not support key passphrases")
        # If key_file is not given, but key_content, write that to a temporary file
        if key_file is None:
            if key_content is None:
                raise KeyParsingError(
                    "one of key_file and key_content must be specified"
                )
            fd, tmpsrc = tempfile.mkstemp()
            self.module.add_cleanup_file(tmpsrc)  # Ansible will delete the file on exit
            f = os.fdopen(fd, "wb")
            try:
                f.write(key_content.encode("utf-8"))
                key_file = tmpsrc
            except Exception as err:
                try:
                    f.close()
                except Exception:
                    pass
                raise KeyParsingError(
                    f"failed to create temporary content file: {err}",
                    exception=traceback.format_exc(),
                ) from err
            f.close()
        # Parse key
        account_key_type = None
        with open(key_file, "r", encoding="utf-8") as fi:
            for line in fi:
                m = re.match(
                    r"^\s*-{5,}BEGIN\s+(EC|RSA)\s+PRIVATE\s+KEY-{5,}\s*$", line
                )
                if m is not None:
                    account_key_type = m.group(1).lower()
                    break
        if account_key_type is None:
            # This happens for example if openssl_privatekey created this key
            # (as opposed to the OpenSSL binary). For now, we assume this is
            # an RSA key.
            # FIXME: add some kind of auto-detection
            account_key_type = "rsa"
        if account_key_type not in ("rsa", "ec"):
            raise KeyParsingError(f'unknown key type "{account_key_type}"')

        openssl_keydump_cmd = [
            self.openssl_binary,
            account_key_type,
            "-in",
            str(key_file),
            "-noout",
            "-text",
        ]
        rc, out, stderr = self.module.run_command(
            openssl_keydump_cmd,
            check_rc=False,
            environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
        )
        if rc != 0:
            raise BackendException(
                f"Error while running {' '.join(openssl_keydump_cmd)}: {stderr}"
            )

        out_text = to_text(out, errors="surrogate_or_strict")

        if account_key_type == "rsa":
            return _extract_rsa_key(
                out_text, key_file=str(key_file) if key_file is not None else None
            )
        if account_key_type == "ec":
            return _extract_ec_key(
                out_text, key_file=str(key_file) if key_file is not None else None
            )
        raise KeyParsingError(
            f"Internal error: unexpected account_key_type = {account_key_type!r}"
        )

    def sign(
        self, *, payload64: str, protected64: str, key_data: dict[str, t.Any]
    ) -> dict[str, t.Any]:
        sign_payload = f"{protected64}.{payload64}".encode("utf8")
        if key_data["type"] == "hmac":
            hex_key = (
                binascii.hexlify(base64.urlsafe_b64decode(key_data["jwk"]["k"]))
            ).decode("ascii")
            cmd_postfix = [
                "-mac",
                "hmac",
                "-macopt",
                f"hexkey:{hex_key}",
                "-binary",
            ]
        else:
            cmd_postfix = ["-sign", key_data["key_file"]]
        openssl_sign_cmd = [
            self.openssl_binary,
            "dgst",
            f"-{key_data['hash']}",
        ] + cmd_postfix

        out: bytes | str

        rc, out, err = self.module.run_command(
            openssl_sign_cmd,
            data=sign_payload,
            check_rc=False,
            binary_data=True,
            environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
        )
        if rc != 0:
            raise BackendException(
                f"Error while running {' '.join(openssl_sign_cmd)}: {err}"
            )

        if key_data["type"] == "ec":
            dummy, der_out, dummy2 = self.module.run_command(
                [self.openssl_binary, "asn1parse", "-inform", "DER"],
                data=out,
                binary_data=True,
                environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
            )
            expected_len = 2 * key_data["point_size"]
            sig = re.findall(
                rf"prim:\s+INTEGER\s+:([0-9A-F]{{1,{expected_len}}})\n",
                to_text(der_out, errors="surrogate_or_strict"),
            )
            if len(sig) != 2:
                der_output = to_text(der_out, errors="surrogate_or_strict")
                raise BackendException(
                    f"failed to generate Elliptic Curve signature; cannot parse DER output: {der_output}"
                )
            sig[0] = (expected_len - len(sig[0])) * "0" + sig[0]
            sig[1] = (expected_len - len(sig[1])) * "0" + sig[1]
            out = binascii.unhexlify(sig[0]) + binascii.unhexlify(sig[1])

        return {
            "protected": protected64,
            "payload": payload64,
            "signature": nopad_b64(to_bytes(out)),
        }

    def create_mac_key(self, *, alg: str, key: str) -> dict[str, t.Any]:
        """Create a MAC key."""
        if alg == "HS256":
            hashalg = "sha256"
            hashbytes = 32
        elif alg == "HS384":
            hashalg = "sha384"
            hashbytes = 48
        elif alg == "HS512":
            hashalg = "sha512"
            hashbytes = 64
        else:
            raise BackendException(
                f"Unsupported MAC key algorithm for OpenSSL backend: {alg}"
            )
        key_bytes = base64.urlsafe_b64decode(key)
        if len(key_bytes) < hashbytes:
            raise BackendException(
                f"{alg} key must be at least {hashbytes} bytes long (after Base64 decoding)"
            )
        return {
            "type": "hmac",
            "alg": alg,
            "jwk": {
                "kty": "oct",
                "k": key,
            },
            "hash": hashalg,
        }

    @staticmethod
    def _normalize_ip(ip: str) -> str:
        try:
            return ipaddress.ip_address(ip).compressed
        except ValueError:
            # We do not want to error out on something IPAddress() cannot parse
            return ip

    def get_ordered_csr_identifiers(
        self,
        *,
        csr_filename: str | os.PathLike | None = None,
        csr_content: str | bytes | None = None,
    ) -> list[tuple[str, str]]:
        """
        Return a list of requested identifiers (CN and SANs) for the CSR.
        Each identifier is a pair (type, identifier), where type is either
        'dns' or 'ip'.

        The list is deduplicated, and if a CNAME is present, it will be returned
        as the first element in the result.
        """
        filename = csr_filename
        data = None
        if csr_content is not None:
            filename = "/dev/stdin"
            data = to_bytes(csr_content)

        openssl_csr_cmd = [
            self.openssl_binary,
            "req",
            "-in",
            str(filename),
            "-noout",
            "-text",
        ]
        rc, out, err = self.module.run_command(
            openssl_csr_cmd,
            data=data,
            check_rc=False,
            binary_data=True,
            environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
        )
        if rc != 0:
            raise BackendException(
                f"Error while running {' '.join(openssl_csr_cmd)}: {err}"
            )

        identifiers = set()
        result = []

        def add_identifier(identifier: tuple[str, str]) -> None:
            if identifier in identifiers:
                return
            identifiers.add(identifier)
            result.append(identifier)

        common_name = re.search(
            r"Subject:.* CN\s?=\s?([^\s,;/]+)",
            to_text(out, errors="surrogate_or_strict"),
        )
        if common_name is not None:
            add_identifier(("dns", common_name.group(1)))
        subject_alt_names = re.search(
            r"X509v3 Subject Alternative Name: (?:critical)?\n +([^\n]+)\n",
            to_text(out, errors="surrogate_or_strict"),
            re.MULTILINE | re.DOTALL,
        )
        if subject_alt_names is not None:
            for san in subject_alt_names.group(1).split(", "):
                if san.lower().startswith("dns:"):
                    add_identifier(("dns", san[4:]))
                elif san.lower().startswith("ip:"):
                    add_identifier(("ip", self._normalize_ip(san[3:])))
                elif san.lower().startswith("ip address:"):
                    add_identifier(("ip", self._normalize_ip(san[11:])))
                else:
                    raise BackendException(f'Found unsupported SAN identifier "{san}"')
        return result

    def get_csr_identifiers(
        self,
        *,
        csr_filename: str | os.PathLike | None = None,
        csr_content: str | bytes | None = None,
    ) -> set[tuple[str, str]]:
        """
        Return a set of requested identifiers (CN and SANs) for the CSR.
        Each identifier is a pair (type, identifier), where type is either
        'dns' or 'ip'.
        """
        return set(
            self.get_ordered_csr_identifiers(
                csr_filename=csr_filename, csr_content=csr_content
            )
        )

    def get_cert_days(
        self,
        *,
        cert_filename: str | os.PathLike | None = None,
        cert_content: str | bytes | None = None,
        now: datetime.datetime | None = None,
    ) -> int:
        """
        Return the days the certificate in cert_filename remains valid and -1
        if the file was not found. If cert_filename contains more than one
        certificate, only the first one will be considered.

        If now is not specified, datetime.datetime.now() is used.
        """
        filename = cert_filename
        data = None
        if cert_content is not None:
            filename = "/dev/stdin"
            data = to_bytes(cert_content)
            cert_filename_suffix = ""
        elif cert_filename is not None:
            if not os.path.exists(cert_filename):
                return -1
            cert_filename_suffix = f" in {cert_filename}"
        else:
            return -1

        openssl_cert_cmd = [
            self.openssl_binary,
            "x509",
            "-in",
            str(filename),
            "-noout",
            "-text",
        ]
        rc, out, err = self.module.run_command(
            openssl_cert_cmd,
            data=data,
            check_rc=False,
            binary_data=True,
            environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
        )
        if rc != 0:
            raise BackendException(
                f"Error while running {' '.join(openssl_cert_cmd)}: {err}"
            )

        out_text = to_text(out, errors="surrogate_or_strict")
        not_after = _extract_date(
            out_text, name="Not After", cert_filename_suffix=cert_filename_suffix
        )
        if now is None:
            now = self.get_now()
        else:
            now = ensure_utc_timezone(now)
        return (not_after - now).days

    def create_chain_matcher(self, *, criterium: Criterium) -> t.NoReturn:
        """
        Given a Criterium object, creates a ChainMatcher object.
        """
        raise BackendException(
            'Alternate chain matching can only be used with the "cryptography" backend.'
        )

    def get_cert_information(
        self,
        *,
        cert_filename: str | os.PathLike | None = None,
        cert_content: str | bytes | None = None,
    ) -> CertificateInformation:
        """
        Return some information on a X.509 certificate as a CertificateInformation object.
        """
        filename = cert_filename
        data = None
        if cert_filename is not None:
            cert_filename_suffix = f" in {cert_filename}"
        else:
            filename = "/dev/stdin"
            data = to_bytes(cert_content)
            cert_filename_suffix = ""

        openssl_cert_cmd = [
            self.openssl_binary,
            "x509",
            "-in",
            str(filename),
            "-noout",
            "-text",
        ]
        rc, out, err = self.module.run_command(
            openssl_cert_cmd,
            data=data,
            check_rc=False,
            binary_data=True,
            environ_update=_OPENSSL_ENVIRONMENT_UPDATE,
        )
        if rc != 0:
            raise BackendException(
                f"Error while running {' '.join(openssl_cert_cmd)}: {err}"
            )

        out_text = to_text(out, errors="surrogate_or_strict")

        not_after = _extract_date(
            out_text, name="Not After", cert_filename_suffix=cert_filename_suffix
        )
        not_before = _extract_date(
            out_text, name="Not Before", cert_filename_suffix=cert_filename_suffix
        )

        sn = re.search(
            r" Serial Number: ([0-9]+)",
            to_text(out, errors="surrogate_or_strict"),
            re.MULTILINE | re.DOTALL,
        )
        if sn:
            serial = int(sn.group(1))
        else:
            serial = convert_bytes_to_int(
                _extract_octets(out_text, name="Serial Number", required=True)
            )

        ski = _extract_octets(
            out_text, name="X509v3 Subject Key Identifier", required=False
        )
        aki = _extract_octets(
            out_text,
            name="X509v3 Authority Key Identifier",
            required=False,
            potential_prefixes=["keyid:", ""],
        )

        return CertificateInformation(
            not_valid_after=not_after,
            not_valid_before=not_before,
            serial_number=serial,
            subject_key_identifier=ski,
            authority_key_identifier=aki,
        )


__all__ = ("OpenSSLCLIBackend",)



#
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

"""
The vlan_parser filter plugin
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
name: vlan_parser
author: Steve Dodd (@idahood)
version_added: "1.0.0"
short_description: The vlan_parser filter plugin.
description:
  - The filter plugin converts a list of vlans to IOS like vlan configuration.
  - Converts list to a list of range of numbers into multiple lists.
  - C(vlans_data | ansible.netcommon.vlan_parser(first_line_len = 20, other_line_len=20))
notes:
  - The filter plugin extends vlans when data provided in range or comma separated.
options:
  data:
    description:
    - This option represents a list containing vlans.
    type: list
    required: True
  first_line_len:
    description:
    - The first line of the list can be first_line_len characters long.
    type: int
    default: 48
  other_line_len:
    description:
    - The subsequent list lines can be other_line_len characters.
    type: int
    default: 44
"""

EXAMPLES = r"""
# Using vlan_parser

- name: Setting host facts for vlan_parser filter plugin
  ansible.builtin.set_fact:
    vlans:
      [
        100,
        1688,
        3002,
        3003,
        3004,
        3005,
        3102,
        3103,
        3104,
        3105,
        3802,
        3900,
        3998,
        3999,
      ]

- name: Invoke vlan_parser filter plugin
  ansible.builtin.set_fact:
    vlans_ranges: "{{ vlans | ansible.netcommon.vlan_parser(first_line_len = 20, other_line_len=20) }}"


# Task Output
# -----------
#
# TASK [Setting host facts for vlan_parser filter plugin]
# ok: [host] => changed=false
#   ansible_facts:
#     vlans:
#     - 100
#     - 1688
#     - 3002
#     - 3003
#     - 3004
#     - 3005
#     - 3102
#     - 3103
#     - 3104
#     - 3105
#     - 3802
#     - 3900
#     - 3998
#     - 3999

# TASK [Invoke vlan_parser filter plugin]
# ok: [host] => changed=false
#   ansible_facts:
#     msg:
#     - 100,1688,3002-3005
#     - 3102-3105,3802,3900
#     - 3998,3999
"""

from ansible.errors import AnsibleFilterError
from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)

from ansible_collections.ansible.netcommon.plugins.plugin_utils.argspec_filter_utils import (
    convert_to_native,
)
from ansible_collections.ansible.netcommon.plugins.plugin_utils.vlan_parser import vlan_parser


try:
    from jinja2.filters import pass_environment
except ImportError:
    from jinja2.filters import environmentfilter as pass_environment


@pass_environment
def _vlan_parser(*args, **kwargs):
    """Extend vlan data"""
    # Extract filter arguments (skip environment which is first arg)
    filter_args = args[1:] if args else []
    keys = ["data", "first_line_len", "other_line_len"]
    data = dict(zip(keys, filter_args))
    data.update(kwargs)

    # Convert to native Python types so ArgumentSpecValidator can deepcopy (Ansible 2.19+)
    data = convert_to_native(data)

    aav = AnsibleArgSpecValidator(data=data, schema=DOCUMENTATION, name="vlan_parser")
    valid, errors, updated_data = aav.validate()
    if not valid:
        raise AnsibleFilterError(errors)
    return vlan_parser(**updated_data)


class FilterModule(object):
    """vlan_parser"""

    def filters(self):
        """a mapping of filter names to functions"""
        return {"vlan_parser": _vlan_parser}
